{
  Promathika - ein multifunktionales Programm zum Zeichnen von Funktionen

    Copyright (C) 2009-2016  Arthur Schmidt, Alexander Staidl

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License Version 3 as published by
    the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
}


unit u_prm_kernel;


//Version: 2.3.83.110 vom 16.08.2016

{$mode delphi}{$H+}

interface

uses math, strutils, Sysutils, dialogs; //dialogs ist nur zum debuggen wichtig; lclintf nur fuer Zeitmessung wichtig.

type
  GNZTyp = array of dword;                      // Grosse-Natuerliche-Zahlen-Typ; qword: 0..18466744073709551615; dword
  GRaZTyp = record                              // Fuer Brueche... ...sollte klar sein.
    Nenner, Zaehler: GNZTyp;
    Negativ: boolean;                           // Negative Zahlen werden unterstützt!
  end;
  GNZListeTyp  = array of GNZTyp;               // Typ für Listen natürlicher Zahlen (z.B. Primzahllisten)
  GRaZListeTyp = array of GRaZTyp;              // Typ für Listen rationaler Zahlen
  PTermTyp = ^TermTyp;
  TermTyp = record
    Opera: cardinal;                            //Operationen (Add, Mul, aber auch Zahl, Variable)
    Arg1, Arg2: PTermTyp;                       //Pointer auf Argumente des Operationen (z.B. Summand1, Summand2)
    Varname: string;                            //Enthaelt Variablennamen im Fall von Opera=002 (Variable)
    ZahlGRaZ: GRaZTyp;                          //Fuer beliebig genaue Rechnung mit rationalen Zahlen
    GlKoZa:double;                              //Fuer Gleitkommazahlen
    Konstante:boolean;                          //Ist true, wenn Opera=002 und Varname=_pi, _e, _alle
  end;
  PTermTypArray = array of PTermTyp;             //Hiermit kann eine Funktion gleichzeitig mehrere Terme übermitteln. Verwendet in den Lösungsroutinen bei mehreren Lösungen.

  tAnalyse = class
    private
      Err:integer;                              // Siehe Datei Error-Codes
      f_PTermTy:PTermTyp;                       // Für den public-Teil: Enthält den Term für f(x)
      GNZ_GlobZahlenbasis:dword;                // Die dem GNZTyp zugrundeliegende Zahlenbasis (Binaersystem=2, Dezimal=10, Hexadezimal=16, u.s.w.); Wird im constructor create festgelegt.
      GNZ_Karazubagrenze: dword;                // Karazubagrenze der Multiplikation

      //HILFSROUTINEN
      function WasNun_Abl(funktion:string):integer;                // Wie WasNun_Funkw, nur auf Ableitungen spezialisiert
      function IstBuchstabe(z:string):boolean;                     // Prüft, ob z ein Buchstabe ist oder nicht.
      function ZeichenErlaubt(z:string):boolean;                   // Prüft, ob in z ausschließlich erlaubte Zeichen verwendet werden.
      function IstZahl(z:string):boolean;                          // Prüft, ob z eine Zahl ist. :Benötigt überall
      function IstNatuerlicheZahl(Eingabe:string):boolean;         // Prüft, ob Eingabe eine natürliche Zahl ist
      function IstVerknuepfung(s:string):boolean;                  // Prüft, ob s aus {+,-,*,/,^}
      function TermKorrPruef(Term:string):string;                  // Prüft, ob die Syntax des vom Benutzer eingegebenen Terms richtig ist.
      procedure ZeichenPruef_FWert(Term:string);                   // Prüft, ob Term für FunkWert geeignet.
      procedure TTTermFreigabe(var Pointer:PTermtyp);              // Gibt den Speicher des Terms frei, auf den der Pointer zeigt.
      function TTKopieVon(Term:PTermTyp):PTermTyp;                 // Erstellt eine komplette Kopie des Terms und gibt den Pointer auf die Kopie zurück

      
      //ZAHLENTHEORIE
      function IstPrim(zahl:int64):boolean;                        // Wahr, wenn zahl Primzahl
      function PrimDavor(zahl:int64):int64;                        // Gibt vorhergehende Prim zurück; RMWdh muss noch integriert werden
      function PrimDanach(zahl:int64):int64;                       // Gibt nachfolgende Prim zurück; RMWdh muss noch integriert werden
      function ggT(a,b:int64):int64;                               // Gibt den größten gemeinsamen Teiler zurück
      function kgV(a,b:int64):int64;                               // Gibt das kleinste gemeinsames Vielfaches zurück
      function GNZadd(a,b:GNZTyp):GNZTyp;                          // Gibt die Summe a + b zurueck. Rechnet mit Grossen Natuerlichen Zahlen (GNZ); Zbasis=Zahlenbasis (Dezimalsystem:Zbasis=10)
      function GNZsub(Minuent,Subtrahend:GNZTyp):GNZTyp;           // WICHTIG: Es muss Minuend>Subtrahend sein!; Rechnet Minuend-Subtrahend.
      function GNZmul(a,b:GNZTyp):GNZTyp;                          // Gibt das Produkt a*b zurueck. Rechnet mit grossen natuerlichen Zahlen.
      function GNZmulword(a:dword;b:GNZTyp):GNZTyp;                // Multipliziert ein dWord mit einem GNZTyp
      function GNZakleinerb(a,b:GNZTyp):boolean; overload;         // Prueft, ob a kleiner ist als b
      function GNZakleinerb(a:GNZTyp; b:dword):boolean; overload;  // Prueft, ob a kleiner ist als b - b:dword;
      function GNZakleinerb(a:dword; b:GNZTyp):boolean; overload;  // Prueft, ob a kleiner ist als b - a:dword;
      function GNZagleichb(a,b:GNZTyp):boolean; overload;          // Prueft, ob sich a und b gleichen.
      function GNZagleichb(a:GNZTyp; b:dword):boolean; overload;   // Prueft, ob sich a und b gleichen - b : dword;
      function GNZeins:GNZTyp;                                     // Das Einselement des GNZ-Typs
      function GNZnull:GNZTyp;                                     // Das Nullelement des GNZ-Typs
      function GNZdiv(Divident, Divisor:GNZTyp):GNZTyp;            // dividiert Divident durch Divisor und gibt das Ergebnis zurueck
      function GNZmod(Divident, Divisor:GNZTyp):GNZTyp;            // Gibt Divident mod Divisor zurueck
      function GNZggt(a,b:GNZTyp):GNZTyp;                          // Wie ggt, nur fuer grosse natuerliche Zahlen (GNZ)
      function GNZkgv(a,b:GNZTyp):GNZTyp;                          // Wie kgv, nur fuer GNZ
      function GNZIstGerade(zahl:GNZTyp):boolean;                  // Gibt zurueck, ob zahl mod 2 = 0
      function GNZPotenz(Basis,Exponent:GNZTyp):GNZTyp;            // Gibt Zahl^Exponent zurueck. Nach einem im Internet gefundenen Algorithmus
      function GNZPotenzModulo(Basis,Exponent,Modulo:GNZTyp):GNZTyp;//Gibt Zahl^Exponent mod Modulo zurueck.
      function GNZFakultaet(nFak:dword):GNZTyp;                    // Gibt n! zurueck
      function GNZNueberK(N,K:dword):GNZTyp;                       // Berechnet N über K (Binomialverteilung).
      function GNZIstPrim(Zahl:GNZTyp):boolean;                    // Wahr, wenn zahl Primzahl
      function GNZPrimDanach(Zahl:GNZTyp):GNZTyp;                  // Gibt die naechste Prinzahl nach Zahl zurueck: Primzahl > Zahl
      function GNZPrimDavor(Zahl:GNZTyp):GNZTyp;                   // Gibt die naechste Primzahl vor Zahl zurueck: Primzahl < Zahl
      function GNZPrimzahlListe(VonGNZ, BisGNZ:GNZTyp):GNZListeTyp;// Gibt eine Liste aller Primzahlen zurück, die zwischen den angegebenen Zahlen liegen.
      function GNZPrimFaktZerl(Zahl:GNZTyp):PTermTyp;              // Gibt als Term das Produkt aus Primfaktoren zurück.
      function GNZMillerRabinPrim(zahl,zeuge:GNZTyp):boolean;      // Zeuge ist zufällige Zahl kleiner zahl-1. Gibt true zurueck, falls zahl ein Primzahl-Kandidat ist.
      function GNZMillerRabinItera(zahl:GNZTyp; itera:dword):boolean;  // Fuehrt den Miller-Rabin-Primzahltest mit itera zufaellig gewaehlten Zeugen durch. Wahrscheinlichkeit, dass zahl nicht prim ist, ist bei Rueckgabe von true (1/4)^itera
      function GNZRandom(OGrenze:GNZTyp):GNZTyp;                   // Gibt eine zufällige Zahl kleiner OGrenze zurueck.
      function GNZZBasisWechsel(QBasisWrd,ZBasisWrd:dword;Zahl:GNZTyp):GNZTyp;  //Wechselt die Basis des Zahlensystems der Zahl, z.B. von Dezimalsystem zu Binaersystem.
      function GNZBasisDezToGlobB(Zahl:GNZTyp):GNZTyp;             // Wechselt die Basis einer Zahl von Dezimalsystem hin zur Globalbasis
      function GNZGlobBToBasisDez(Zahl:GNZTyp):GNZTyp;             // Wechselt die Basis einer Zahl von Globalbasis hin zum Dezimalsystem
      function WordToGNZTyp(WZahl:dword):GNZTyp;                   // Wandelt eine dword-Zahl in eine Zahl vom GNZTyp um.

      function GRaZKuerzen(Bruch:GRaZTyp):GRaZTyp;                 // Kuerzt den Bruch
      function GRaZadd(a,b:GRazTyp):GRaZTyp;                       // Rechnet a+b und kuerzt die Summe anschliessend
      function GRazsub(Minuent,Subtrahend:GRaZTyp):GRaZTyp;        // Rechnet a-b und kuerzt die Summe anschliessend
      function GRaZmul(a,b:GRaZTyp):GRaZTyp;                       // Multipliziert die Brueche miteinander
      function GRaZdiv(Divident,Divisor:GRaZTyp):GRaZTyp;          // Dividiert durch Multiplikation mit dem Kehrwert
      function GRaZakleinerbBetrag(a,b:GRaZTyp):boolean;           // Gibt true zurueck, wenn |a| kleiner |b|
      function GRaZakleinerb(a,b:GRaZTyp):boolean;                 // Gibt true zurueck, wenn a kleiner b
      function GRaZagleichbBetrag(a,b:GRaZTyp):boolean;            // Gibt true zurueck, wenn |a| gleich |b|
      function GRaZagleichb(a,b:GRaZTyp):boolean;                  // Gibt true zurueck, wenn a gleich b
      function GRaZeins:GRaZTyp;                                   // Gibt 1 zurueck.
      function GRaZnull:GRaZTyp;                                   // Gibt 0 zurueck
      function GRaZcopy(z:GRaZTyp):GRaZTyp;                        // Muss zum Kopieren zweier GRaZ verwendet werden, da ansonsten nur die Array-Pointer kopiert werden.
      function WordToGRaZTyp(Zaehler,Nenner:dword;Negativ:boolean):GRaZTyp;// Wandelt 2 dwords und ein boolean in eine rationale Zahl GRaZTyp um.
      
      //FUNKWERT
      function FunkWert_NM(Pointer:PTermTyp; XWert:double):double; // Neue Methode - Intern: Rueckgabe des Funktionswertes an Stelle XWert. Pointer muss bei Aufruf 0 sein.
      function UnbekannteErsetzen(Term, Unbekannte, Ersatz: PTermTyp):PTermTyp; // Zurückgegeben wird ein Term, in dem Unbekannte durch Ersatz ersetzt wurde. Für TTLoeseGleichung, TTaylorEntwicklung benötigt.
      procedure Dezimalisierung(var Pointer:PTermTyp);             // Zieht in einem Term alles, was nicht mit einer Variablen verknüpft ist, zu einer GlKoZa zusammen.

      //ANALYSIS
      function TTAbleitung(Funktion,Variable:PTermTyp):PTermTyp;   // Gibt Pointer auf die Ableitung von funktion zurueck. Die Variable, nach der abgeleitet wird, kann auch eine Funktion sein. So kann man z.B. (sin(x))^2 nach sin(x) ableiten.
      function TTTaylorEntwicklung(Term, Variable, EntwPunkt:PTermTyp; Potenz:dword):PTermTyp; // Führt die Taylor-Entwicklung durch.

      //PARSER
      function StrToTermtyp(Eingabe:string):PTermTyp;              //Wandelt den Term vom Typ String nach Termtyp um und gibt Pointer darauf zurueck
      function TermtypToStrRek(Pointer:PTermTyp;DezimalbruchZiffAnz:word):string;  //Eigenständige Routine, die von TermtypToStr aufgerufen wird.
      function TermtypToStr(var Pointer:PTermTyp;DezimalbruchZiffAnz:word):String; //Wandelt einen Term vom Typ TermTyp in einen String um. DezimalbruchZiffAnz gibt die Genauigkeit der zurückgegebenen Zahlen wieder:
                                                                                   // 0: Exakte Ausgabe - Zahlen werden in Brüchen zurückgegeben, u.s.w.; 1-16: Es werden Dezimalbrüche (Kommazahlen) mit der entsprechenden Ziffernanzahl zurückgegeben.
      function StrToGNZTyp(Eingabe:string):GNZTyp;                 //Wandelt eine natuerliche Zahl in string-Forum um in den Typ GNZTyp von Basis 10. Eingabe darf nur Zahlen enthalten.
      function GNZTypToStr(Eingabe:GNZTyp):string;                 //"Umkehrfunktion" zu StrToGNZTypDez;
      function GRaZTypToDouble(Eingabe:GRaZTyp):double;            //Wandelt eine Zahl vom GRaZTyp in eine Zahl vom Typ double um.
      function DoubleToStr(Eingabe:double;ZifferAnz:word):string;  //Wandelt eine double in einen String mit der Genauigkeit ZifferAnz und ersetzt '5E-10' durch '5*10^(-10)'
      function GRaZTypToStr(Eingabe:GRaZTyp;ZifferAnz:word):string;//Wandelt eine Zahl vom GRaZTyp in einen String um. Die zurückgegebene Zahl hat dabei (maximal) die angegebene Anzahl an Ziffern.

      //ALGEBRA
      procedure TTkommu(Term:PTermTyp);                            //Wendet, wenn Opera=4,5, das Kommutativgesetz der Add/Mult an.
      procedure TTassoz(Term:PTermTyp;ArgNr:Byte);                 //Wendet, wenn Opera=4,5, das Assoziativgesetz der Add/Mult an; ArgNr=1:(a+b)+c --> a+(b+c); ArgNr=2:a+(b+c) --> (a+b)+c
      procedure TTdistri(Term:PTermTyp;ArgNr:Byte);                //Wendet das Distributivgesetz an. ArgNr=1: Klammer bei Arg1 wird aufgeloest; ArgNr=2: Klammer bei Arg2 wird aufgeloest.
      procedure TTVereinfachung(var Term:PTermTyp);                //Sortiert den Term nach bestimmten Kriterien um und vereinfacht ihn anschließend.
      procedure TTTrigoSymVereinfachung(var Term:PTermTyp; punktsymmetrisch:boolean); // Vereinfacht die trigonometrischen Funktionen durch Symmetrieeigenschaften. Bsp.: sin(-1*x)=-1*sin(x)
                                                                                      // punktsymmetrisch=false, falls achsensymmetrisch.
      procedure TTWurzelTeilwZiehen(var Term:PTermTyp);            //Zieht die Wurzel teilweise. Argument muss Potenz mit GRaZ als Basis und ZahlMitNenner als Exponent sein.
      function  TTakleinerb(a,b:PTermTyp):boolean;                 //Vergleicht zwei Terme und definiert eine Ordnung.
      function  TTagleichb(a,b:PTermTyp):boolean;                  //Gibt true zurück, wenn a=b. Term muss sortiert sein.
      procedure TTEntwicklung(var Term:PTermTyp);                  //Bringt einen Term durch Ausmultiplizieren in eine eindeutige Schreibweise.
      procedure TTEinzelklammernAufloesen(var Term:PTermTyp);      //Wie TTEntwicklung, nur werden jetzt alle Einzelklammern, wie a*(b+c) ausmultipliziert, (b+c)^n bleibt aber erhalten. Benötigt für TTAbleitung.
      function  TTLoeseGleichung(LinkeSeite, RechteSeite, Unbekannte:PTermTyp):PTermTypArray;
                                                                   //Löst die Gleichung "LinkeSeite = RechteSeite" nach "Unbekannte" auf.
      function  TTTermToSummandForm(var Term:PTermTyp):PTermTypArray; //Multipliziert den Term aus und gibt seine Summanden in einem PTermTypArray zurück. Dieses vereinfacht viele Gleichungslösungsverfahren.
      procedure TTBruchEntferner(var SummandForm:PTermTypArray; Unbekannte:PTermTyp);
                                                                   //Löst alle Brüche auf, indem die Summanden mit jeden Brüchen multipliziert werden, welche die Unbekannte enthalten.
      function  TTUnbekannteEnthalten(Term,Unbekannte:PTermTyp):boolean;   //Gibt true zurück, wenn die Unbekannte im Term enthalten ist.
      function  TTKeineVariableEnthalten(Term:PTermTyp):boolean;   //Gibt true zurück, wenn der Term keine Variable enthält. D.h. er besteht nur aus Zahlenverknüpfungen.
      function  TTPotenzLoeser(var SummandForm:PTermTypArray; Unbekannte:PTermTyp):PTermTypArray;
                                                                   //Löst Potenzgleichungen und gibt die Lösungen in einem Array zurück.


    public
      constructor create;
      destructor free;

      //PUBLIC
      procedure FEing(funktion:string);                            // Eingabe der Funktion für FunkWert: f_FWert
      procedure FWert_polar(phi:double; var r,x,y:double);     // Gibt zum Winkel Phi den Radius r sowie die Position des Punktes in kart. Koord. (x,y) zurueck
      function  FWert(Xwert:double):double;                    // Gibt Funktionswert an der Stelle XWert zurück.
      function  Error:integer;                                     //Intern: Behebung von Darstellungsfehlern
      procedure ResetError;                                        // Setzt Err:=0;

      //Schnittstelle für die Konsole
      function SAbleitung(funktionStr:string; VariableStr:string):string; overload;
                                                                            // Schnittstelle nach außen für Ableitungen
      function SAbleitung(funktionSTr:string; VariableStr:string; VielfachheitStr:string):string; overload;
                                                                            // Schnittstelle für n-fache Ableitung.
      function SBasiswechsel(QBasisStr,ZBasisStr,ZahlStr:string):string;    // Schnittstelle für GNZZBasisWechsel
      function SDiv(DividentStr,DivisorStr:string):string;                  // Schnittstelle für GNZmod
      function SEntwicklung(TermStr:string):string;                         // Schnittstelle für Termentwicklung
      function SFakultaet(ZahlSTr:string):string;                           // Schnittstelle für GNZFakultaet
      function SNueberK(NStr,KStr:string):string;                           // Schnittstelle für GNZNueberK
      function Sggt(aStr,bStr:string):string;                               // Schnittstelle für GGT (größter gemeinsamer Teiler)
      function SIstPrim(ZahlStr:string):string ;                            // Schnittstelle für GNZIstPrim
      function Skgv(aStr,bStr:string):string;                               // Schnittstelle für KGV (kleinstes gemeinsames Vielfaches)
      function SLoeseGleichung(GleichungStr,UnbekannteStr:string):string;   // Schnittstelle für TTLoeseGleichung
      function SMillerRabinPrim(zahlStr,zeugeStr:string):string;            // Schnittstelle für GNZMillerRabinPrim
      function SMillerRabinItera(zahlStr,iteraStr:string):string;           // Schnittstelle für GNZMillerRabinItera
      function SMod(DividentStr,DivisorStr:string):string;                  // Schnittstelle für GNZMod
      function SPrimDanach(ZahlStr:string):string;                          // Schnittstelle für GNZPrimDanach
      function SPrimDavor(ZahlStr:string):string;                           // Schnittstelle für GNZPrimDavor
      function SPrimFaktZerl(ZahlStr:string):string;                        // Schnittstelle für GNZPrimFaktZerl
      function SPrimzahlListe(VonStr,BisStr:string):string;                 // Schnittstelle für GNZPrimzahlListe
      function SPotenzModulo(basisStr,exponentStr,moduloStr:string):string; // Schnittstelle für GNZPotenzModulo
      function STaylorEntwicklung(TermStr, VariableStr, EntwPunktStr, PotenzStr:string):string;
                                                                            // Schnittstelle für TTTaylorEntwicklung
      function SVereinfachung(TermStr:string; DezimalbruchZiffernAnz:word):string;
                                                                            // Schnittstelle für Termvereinfachung
      function SGNZRandom(zahlStr:string):string;                           // Schnittstelle für GNZRandom

  end;

implementation

{$INCLUDE u_prm_kernel_hilfsroutinen.pas}
{$INCLUDE u_prm_kernel_zahlentheorie.pas}
{$INCLUDE u_prm_kernel_funkwert.pas}
{$INCLUDE u_prm_kernel_analysis.pas}
{$INCLUDE u_prm_kernel_parser.pas}
{$INCLUDE u_prm_kernel_algebra.pas}

{$INCLUDE u_prm_kernel_public.pas}


constructor tAnalyse.create;
begin
  inherited create;
  f_PTermTy:=nil;
  DefaultFormatSettings.DecimalSeparator:=',';
  Err:=0;
  GNZ_GlobZahlenbasis:=2147483648;  //10=Dezimalsystem
  GNZ_Karazubagrenze:=44;      //Karazubagrenze der Multiplikation
  Randomize;
end;

destructor tAnalyse.free;
begin
  If assigned(f_PTermTy) then TTTermFreigabe(f_PTermTy);
  inherited destroy;
end;



end.

