{
  Promathika - ein multifunktionales Programm zum Zeichnen von Funktionen

    Copyright (C) 2009  Arthur Schmidt, Alexander Staidl

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License Version 3 as published by
    the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
}


unit uFunclistDB;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Graphics, lclproc,dialogs,u_prm_Settings,FileUtil;

  type tDrawoptions=record
    Preview:boolean;               //ID 4000
    ignoreInts:boolean;            //ID 4001
    ignoreCols:boolean;            //ID 4002
    ignoreCurvWidth:boolean;       //ID 4003
    ignoreUnsharp:boolean;         //ID 4004
    ignorePerf:boolean;            //ID 4005  AA-Modus, -Stufe und Thread-Anzahl
  end;

  type tFuncSet=record
    FuncString:String;             //5000   min 2byte,  nicht für Speicherung als typisierte Datei geeignet
    IntMin:extended;               //5001   10byte
    IntMax:extended;               //5002   10byte
    CurvWidth:double;              //5003    8byte
    CurvCol:tColor;                //5004    4byte
    CurvCol2:tColor;               //5005    4byte
    CurvGrad:LongInt;              //5006    4byte
    unsharpnes:extended;           //5007   10byte
    AA_Level_hor:byte;             //5008    1byte
    AA_Level_ver:byte;             //5009   1byte
    AA_Mode:byte;                  //5010   1byte
    ThreadCount:byte;              //5011   1byte
    Active:boolean;                //5012   1byte
    FuncName:String;           //5013   min 2byte
  end;                             //min. 130byte

  type tStrArr=Array of String;
  type tFunclist=array of tFuncSet;

  type tFunclistDB=class
    private
      List:tFunclist;
      FilepathDB:UTF8string;
      DrwOpt:tDrawOptions;
      procedure StrToOption(OptionLine:UTF8string; var workSet:tFuncSet);
      function openList(Filepath:UTF8string; var Target:tFuncList;var refoptions:tDrawOptions):boolean;
      procedure getDefaults(var RefOptions:tDrawOptions);
      procedure OptionsToStrArr(const RefOptions:tDrawOptions;var thisStrArr:tStrArr);
      procedure FuncSetToStrArr(const RefSet:tFuncSet;var thisStrArr:tStrArr);
    public
      procedure getDefaults(var RefSet:tFuncSet);
      constructor create;
      destructor free;
      procedure SetDrawOptions(Set_Options:tDrawOptions);  //Einstellungen, die ignoriert werden sollen, einstellen
      function GetDrawOptions:tDrawOptions;                //Einstellungen, die ignoriert werden sollen, einlesen
      procedure openList(Filepath:UTF8string);                 //Liste öffen. Gleichzeitig wird neue Ram-DB erzeugt und der FilePathDB auf die Datei gelegt
      procedure integrateList (Filepath:UTF8string);           //Liste in bestehende Liste integrieren
      function saveList (filePath:UTF8string):boolean;                //Liste speichern
      procedure ModifyList (const index:integer;const funcdata:tFuncset); //Hinzufügen einer neuen Funktion, wenn index=-1 oder ändern einer bestehenden
      procedure deletefunc (const DelIndex:LongInt);                     //Funktion aus Ram-DB löschen
      procedure StartNewFile;
      procedure GetDrawList(GlobalSettings:tGlobalSet;var FuncList:tFuncList); //hier wird eine Liste mit ignorierten Settings zurückgegeben
      procedure ListResults(var funcList_im:tFuncList);
      function Ergebnisse:integer;
      function ShowFile:UTF8string;
  end;

//===================================================================================================================================
//===================================================================================================================================

implementation

  constructor tFunclistDB.create;
    begin
      inherited create;
      StartNewFile;
    end;

//===================================================================================================================================

  function tFuncListDB.Ergebnisse:integer;
    begin
      result:=length(List);
    end;

//===================================================================================================================================

  destructor tFunclistDB.free;
    begin
      setlength(List,0);
      List:=nil;
      FilepathDB:='';
      inherited destroy;
    end;

//===================================================================================================================================

  procedure tFuncListDB.SetDrawOptions(Set_Options:tDrawOptions);
    begin
      Drwopt:=Set_Options;
    end;

//===================================================================================================================================

  function tFuncListDb.GetDrawOptions:tDrawOptions;
    begin
      result:=DrwOpt;
    end;

//===================================================================================================================================

  procedure tFunclistDB.StartNewFile;
    begin
      FilePathDB:='';
      setlength(List,0);
      getdefaults(DrwOpt);
    end;

//===================================================================================================================================

  procedure tFuncListDB.getdefaults(var RefSet:tFuncSet);
    begin
      with RefSet do begin
        FuncString:='';             //ID 5000   min 2byte,  nicht für Speicherung als typisierte Datei geeignet
        IntMin:=0;                  //ID 5001   10byte
        IntMax:=0;                  //ID 5002   10byte
        CurvWidth:=1.5;             //ID 5003    8byte
        CurvCol:=clBlue;            //ID 5004    4byte
        CurvCol2:=clNavy;           //ID 5005    4byte
        CurvGrad:=-1;               //ID 5006    4byte
        unsharpnes:=0.1;           //ID 5007   10byte
        AA_Level_hor:=4;            //ID 5008    1byte
        AA_Level_ver:=3;            //ID 5009   1byte
        AA_Mode:=2;                 //ID 5010   1byte
        ThreadCount:=4;             //ID 5011   1byte
        Active:=true;               //ID 5012
        FuncName:='';               //ID 5013
      end;
    end;

//===================================================================================================================================

  procedure tFuncListDB.getdefaults(var RefOptions:tDrawOptions);
    begin
      with RefOptions do begin
        preview:=true;             //ID 4000
        ignoreInts:=false;          //ID 4001
        ignoreCols:=false;          //ID 4002
        ignoreCurvWidth:=false;     //ID 4003
        ignoreUnsharp:=false;       //ID 4004
        ignorePerf:=false;          //ID 4005
      end;
    end;

//===================================================================================================================================

  procedure tFunclistDB.StrToOption(OptionLine:UTF8string; var workSet:tfuncSet);
   var
     OptIDBuffer:longint=-1;
     PosBuf:array[0..2]of longint;
    begin
      If (length(OptionLine)>0) then begin
        PosBuf[0]:=Pos('#',Optionline);  //Abschlusszeichen der ID
        PosBuf[1]:=Pos('|',OptionLine);  //Einleitungszeichen für den Optionswert
        PosBuf[2]:=Pos('@',OptionLine);  //Einleitungszeichen für einen evtl. Kommentar

        If (PosBuf[0]>0)and(PosBuf[1]>PosBuf[0]+1)then begin
          OptIdBuffer:=StrToInt(trim(Copy(OptionLine,1,PosBuf[0]-1)));
          If PosBuf[2]>PosBuf[1]+1 then OptionLine:=trim(copy(OptionLine,PosBuf[1]+1,PosBuf[2]-PosBuf[1]-1)) else
           If PosBuf[2]=0 then OptionLine:=trim(Copy(OptionLine,Posbuf[1]+1,length(OptionLine)-PosBuf[1])) else OptIdBuffer:=-1;

          If (OptIDBuffer>-1) then begin
            Case OptIDBuffer of
              //DrawOptions
              4000: DrwOpt.Preview:=StrToBool(OptionLine);
              4001: DrwOpt.ignoreInts:=StrToBool(OptionLine);
              4002: DrwOpt.ignoreCols:=StrToBool(OptionLine);
              4003: DrwOpt.ignoreCurvWidth:=StrToBool(OptionLine);
              4004: DrwOpt.ignoreUnsharp:=StrToBool(OptionLine);
              4005: DrwOpt.ignorePerf:=StrToBool(OptionLine);
              //FuncSet
              5000: WorkSet.FuncString:=OptionLine;
              5001: WorkSet.IntMin:=StrToFloat(OptionLine);
              5002: WorkSet.IntMax:=StrToFloat(OptionLine);
              5003: WorkSet.CurvWidth:=StrToFloat(OptionLine);
              5004: WorkSet.CurvCol:=StrToInt(OptionLine);
              5005: WorkSet.CurvCol2:=StrToInt(OptionLine);
              5006: WorkSet.CurvGrad:=StrToInt(OptionLine);
              5007: WorkSet.unsharpnes:=StrToFloat(OptionLine);
              5008: WorkSet.AA_Level_hor:=StrToInt(OptionLine);
              5009: WorkSet.AA_Level_ver:=StrToInt(OptionLine);
              5010: WorkSet.AA_Mode:=StrToInt(OptionLine);
              5011: WorkSet.ThreadCount:=StrToInt(OptionLine);
              5012: WorkSet.Active:=StrToBool(OptionLine);
              5013: Workset.FuncName:=OptionLine;
            end;
          end;
        end;
      end;
    end;

//===================================================================================================================================

  procedure tFunclistDB.openList(FilePath:UTF8string);
    begin
      if OpenList(FilePath,List,DrwOpt)=true then FilePathDB:=FilePath;
    end;

//===================================================================================================================================

  function tFuncListDB.openList(Filepath:UTF8string; var Target:tFuncList; var refOptions:tDrawOptions):boolean;
   var
     ListFile:TextFile;
     Setbuf:tFuncSet;
     s:String='';
    begin
      result:=false;
      If FileExistsUTF8(FilePath)then begin
        result:=true;
        try
          AssignFile(ListFile,UTF8ToSys(FilePath));
          {$i-}
          Reset(ListFile);
          {$i+}
          getDefaults(refOptions);
          getDefaults(SetBuf);
          setlength(Target,0);
          while not eof(ListFile)do begin
            readln(ListFile,s);
            s:=Trim(s); //Leerzeichen an Stringenden entfernen
            if s<>'' then begin
              try
                StrToOption(s,SetBuf);
              except

              end;
            end;
            If (s='new function')and(SetBuf.FuncString<>'')then begin
              setlength(List,length(List)+1);
              List[length(List)-1]:=SetBuf;
              getdefaults(SetBuf);
            end;
          end;
          If SetBuf.FuncString<>'' then begin
            setlength(List,length(List)+1);
            List[length(List)-1]:=SetBuf;
          end;
          CloseFile(ListFile);
        except
          CloseFile(ListFile);
          result:=false;
        end;
      end;
    end;

//===================================================================================================================================

  procedure tFuncListDB.integrateList(FilePath:UTF8String);
   var
     IntegrateData:tFuncList;
     opt:tDrawOptions;
     BufData:tFuncSet;
     i,IntegrateSize,ListSize:longInt;
    begin
      If FilePathDB<>'' then begin
        OpenList(FilePath,IntegrateData,opt);
        IntegrateSize:=length(IntegrateData);
        If IntegrateSize > 0 then begin
          ListSize:=length(List);
          Setlength(List,ListSize+IntegrateSize);
          For i:=0 to IntegrateSize-1 do begin
            List[ListSize+i]:=IntegrateData[i];
          end;
          IntegrateData:=nil;
        end;
      end;
    end;

//===================================================================================================================================

  procedure tFuncListDB.OptionsToStrArr(const RefOptions:tDrawOptions;var thisStrArr:tStrArr);
    begin
      setlength(thisStrArr,6);
      thisStrArr[0]:='4000 #'+'Preview| '+BoolToStr(RefOptions.Preview);
      thisStrArr[1]:='4001 #'+'IgnoreInts| '+BoolToStr(RefOptions.ignoreInts);
      thisStrArr[2]:='4002 #'+'IgnoreCols| '+BoolToStr(RefOptions.ignoreCols);
      thisStrArr[3]:='4003 #'+'IgnoreCurvWidths| '+BoolToStr(RefOptions.ignoreCurvWidth);
      thisStrArr[4]:='4004 #'+'Ignore unsharpnes| '+BoolToStr(RefOptions.ignoreUnsharp);
      thisStrArr[5]:='4005 #'+'ignore PerfSettings| '+BoolToStr(RefOptions.ignorePerf);
    end;

//===================================================================================================================================
   procedure tFuncListDB.FuncSetToStrArr(const RefSet:tFuncSet;var thisStrArr:tStrArr);
     begin
       setlength(thisStrArr,14);
       thisStrArr[0]:='5000 #'+'funcString| '+RefSet.FuncString;
       thisStrArr[1]:='5001 #'+'IntMin| '+FloatToStr(RefSet.IntMin);
       thisStrArr[2]:='5002 #'+'IntMax| '+FloatToStr(RefSet.IntMax);
       thisStrArr[3]:='5003 #'+'CurvWidth| '+FloatToStr(RefSet.CurvWidth);
       thisStrArr[4]:='5004 #'+'CurvCol| '+IntToStr(RefSet.CurvCol);
       thisStrArr[5]:='5005 #'+'CurvCol2| '+IntToStr(RefSet.CurvCol2);
       thisStrArr[6]:='5006 #'+'CurvGrad| '+IntToStr(RefSet.CurvGrad);
       thisStrArr[7]:='5007 #'+'Unsharpnes| '+FloatToStr(RefSet.unsharpnes);
       thisStrArr[8]:='5008 #'+'AA_Level_hor| '+IntToStr(RefSet.AA_Level_hor);
       thisStrArr[9]:='5009 #'+'AA_Level_ver| '+IntToStr(RefSet.AA_Level_ver);
       thisStrArr[10]:='5010 #'+'AA_Mode| '+IntToStr(RefSet.AA_Mode);
       thisStrArr[11]:='5011 #'+'ThreadCount| '+IntToStr(RefSet.ThreadCount);
       thisStrArr[12]:='5012 #'+'active| '+BoolToStr(RefSet.Active);
       thisStrArr[13]:='5013 #'+'funcName| '+RefSet.FuncName;
     end;

//===================================================================================================================================

  function tFuncListDB.saveList(FilePath:UTF8String):boolean;
   var
     ListFile: TextFile;
     StrArrBuf:tStrArr;
     i,intBuf,k:integer;
    begin
      result:=false;
      If ((FilePath<>'')or((FilePath='')and(FilePathDB<>'')))and(length(list)>0) then begin
        If (FilePath='')and(FilePathDB<>'')then FilePath:=FilePathDB;
        try
          AssignFile(ListFile,UTF8ToSys(FilePath));
          {$i-}
          ReWrite(ListFile);
          {$i+}
          If length(List)>0 then begin               //Wenn die Datei zusaetzlich nicht ganz leer ist, wird diese ausgelesen
            OptionsToStrArr(DrwOpt,StrArrBuf);
            intBuf:=length(StrArrBuf)-1;
            writeln(ListFile,'options');
            for i:=0 to intbuf do begin
              writeln(ListFile,StrArrBuf[i]);
            end;
            writeln(ListFile,'');
            For k:=0 to length(List)-1 do begin
              writeln(ListFile,'new function');
              FuncSetToStrArr(List[k],StrArrBuf);
              IntBuf:=length(StrArrBuf)-1;
              If IntBuf>-1 then begin
                For i:=0 to IntBuf do begin      //Auslesen der Datei
                  writeln(ListFile,StrArrBuf[i]);
                end;//endfor
                writeln(ListFile,'');
              end;
            end;
            result:=true;//wenn etwas geladen wurde >0
          end;//endIf
          CloseFile(ListFile);
          If FilePathDB='' then FilePathDB:=FilePath;
        except
          CloseFile(ListFile);
        end;
      end;

      If (FilePath='')and(FilePathDB='')then MessageDlg('*Info*','Keine geöffnete Datei gefunden! Versuchen Sie es bitte mit dem Befehl "speichern unter".', mtinformation, [mbOk] , 0);
      if length(list)=0 then MessageDlg('*Info*','Leere Dateien werden nicht gespeichert!', mtinformation, [mbOk] , 0);
    end;

//===================================================================================================================================

  procedure tFuncListDB.deletefunc (const DelIndex:LongInt);
   var
     counter,ListSize:integer;
    begin
      ListSize:=length(List);
      If (DelIndex>-1)and(DelIndex<ListSize-1) then begin
        for Counter:=DelIndex to ListSize-2 do begin
          List[Counter]:=List[Counter+1];
        end;
      end;
      If (DelIndex>-1)and(ListSize>0)and(DelIndex<=ListSize-1)then setlength(List,ListSize-1);
    end;

//===================================================================================================================================

  procedure tFuncListDB.ModifyList (const index:integer;const funcdata:tFuncset);
    begin
      if (Index<length(list))and(Index>-1)and(funcdata.FuncString<>'')then List[index]:=Funcdata else
       if (Index=-1)and(funcdata.FuncString<>'') then begin
         setlength(list,length(list)+1);
         List[length(list)-1]:=funcData;
       end else
         MessageDlg('!Fehler!','Es gibt keine Funktion an vorgegebener Stelle, nichts wird unternommen.', mterror, [mbOk] , 0);
    end;

//===================================================================================================================================

  procedure tFuncListDB.ListResults(var funcList_im:tFuncList);
   var
     i:integer;
    begin
      setlength(funcList_im,length(List));
      for i:= 0 to length(list)-1 do begin
        FuncList_im[i]:=List[i];
      end;
    end;

//===================================================================================================================================

 procedure tFuncListDB.GetDrawList(GlobalSettings:tGlobalSet;var FuncList:tFuncList);
  var
    n,l:longInt;
   begin
     l:=length(List);
     setlength(FuncList,l);
     for n:=0 to l-1 do begin
       FuncList[n]:=List[n];
     end;

     For n:=0 to l-1 do begin
       If DrwOpt.ignorePerf=true then begin
         FuncList[n].AA_Level_hor:=GlobalSettings.PerformanceSet.AA_Level_hor;
         FuncList[n].AA_Level_ver:=GlobalSettings.PerformanceSet.AA_Level_ver;
         FuncList[n].ThreadCount:=GlobalSettings.PerformanceSet.ThreadCount;
         FuncList[n].AA_Mode:=GlobalSettings.PerformanceSet.AA_Mode;
       end;
       If DrwOpt.ignoreUnsharp=true then FuncList[n].unsharpnes:=GlobalSettings.PerformanceSet.unsharpnes;
       if DrwOpt.ignoreInts=true then begin
         FuncList[n].IntMin:=GlobalSettings.Kartesianset.xmin;
         FuncList[n].IntMax:=GlobalSettings.Kartesianset.xmax;
       end;
       If DrwOpt.ignoreCols=true then begin
         FuncList[n].CurvCol:=GlobalSettings.ColorSet.CurvCol;
         FuncList[n].CurvCol2:=GlobalSettings.ColorSet.CurvCol2;
         FuncList[n].CurvGrad:= GlobalSettings.ColorSet.CurvGrad;
       end;
       If DrwOpt.ignoreCurvWidth=true then FuncList[n].CurvWidth:=Globalsettings.PerformanceSet.CurvWidth;
     end;
     //hier können Mods angehängt werden, wie ignorierte Settings
   end;

//===================================================================================================================================

  function tFuncListDB.ShowFile:UTF8string;
    begin
      result:=FilePathDB;
    end;

end.

