{
  Promathika - ein multifunktionales Programm zum Zeichnen von Funktionen

    Copyright (C) 2009  Arthur Schmidt, Alexander Staidl

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License Version 3 as published by
    the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
}


unit uGraphCom;

{$mode delphi}{$H+}

interface

uses
  Classes, SysUtils, Graphics,math, u_prm_kernel, u_prm_Settings,dialogs;

  type twordsarray= array of word;

  type tThreadstats=record
    postprocess:longInt;
    Resettime:longInt;
    RenderTime:longInt;
    parser:longInt;
    StepCount:int64;
  end;

  type twordpoint=record
   x:word;
   y:word;
  end;

  type tPix=record
    Punkt:twordPoint;   //4 byte
    Farbe:tColor;        //4 byte
  end;                   //= 8 byte per Result-Pixel

  type tVec2D = record
    x,y:double;
  end;

  type tPixArray = array of tPix;

  type tPointCloud = array of tVec2D;

  tGraphCom = class(tThread)
  private
    xMinThread,xMaxThread,IndividualOffset:double;    //xMinThread und xMaxThread sind Thread-Abhängig
    ZoomX,ZoomY,IntMin,IntMax:double;// -> nicht thread-Abhängig
    ThreadResX:Word;
    xMinThread_Canv:longInt;
    DrawSet:tGlobalSet;
    funktion:string;
    ana:tAnalyse;
    ThreadNumber,GraphCompMode:byte;
    ThreadIsValid:boolean;
    PixArray:tPixArray;
    PixPlan:array of Array of LongInt;
    AnteilPlan:array of byte;
    Subpixels:Array of Dword;
    QSubpixels:Array of QWord;
    PCloud:tPointCloud;
    stats:tThreadStats;
    function XRealToCanv (const x:double):double;
    function YRealToCanv (const y:double):double;
    procedure C_unbufferedTetragonAACurv(const ComputationMode:byte);
    procedure C_BufferedHexagonAACurv(const ComputationMode:byte);//0=default, 1=NumIntegral, 2=CloudMode
    procedure C_BufferedPointAACurv(const ComputationMode:byte);
    procedure C_BufferedRoundRectAACurv(const ComputationMode:byte);
    procedure ResetPixMap;
    Procedure generate_Results;
    function GetBit(const Value: DWord; const Bit: Byte): Boolean;
    function SetBit(const Value: DWord; const Bit: Byte): DWord;
    function QGetBit(const Value: QWord; const Bit: Byte): Boolean;
    function QSetBit(const Value: QWord; const Bit: Byte): QWord;
    procedure SharedCreate(props:tGlobalSet;TNumber:byte;IntMin_im,IntMax_im:double);
  protected
    procedure Execute; override;
  public
    constructor create(CreateSuspended: boolean;props:tGlobalSet;TNumber:byte;IntMin_im,IntMax_im:double;funcString:string); overload;
    constructor create(CreateSuspended: boolean;props:tGlobalSet;TNumber:byte;IntMin_im,IntMax_im:double;const PointCloud:tPointCloud); overload;
    procedure SetIntegralOffset(offset_imp:double);
    Destructor free;
    function GetPixArray:tPixArray;
    function GetTimes:tThreadStats;
  end;

implementation

uses uKonsole;

//=====================================================================================================================
//=====================================================================================================================
//=====================================================================================================================

 procedure tGraphCom.SharedCreate(props:tGlobalSet;TNumber:byte;IntMin_im,IntMax_im:double);
   begin
     If (IntMin_im<Props.Kartesianset.xmax)and(IntMax_im>Props.Kartesianset.xmin)or(IntMin_im=Intmax_im) then begin
        DrawSet:=Props;
        zoomx:=DrawSet.MainSizing.DrawResX/(DrawSet.Kartesianset.xmax-DrawSet.Kartesianset.xmin);             //notwendige Werte werden initialisiert
        zoomy:=DrawSet.MainSizing.DrawResY/(DrawSet.Kartesianset.ymax-DrawSet.Kartesianset.ymin);
        ThreadNumber:=TNumber;
        IntMin:=IntMin_im; IntMax:=IntMax_im;
        If (IntMin=IntMax) then begin
          IntMin:=DrawSet.Kartesianset.xmin;
          IntMax:=DrawSet.Kartesianset.xmax;
        end;
        If (IntMin<DrawSet.Kartesianset.xmin) then IntMin:=DrawSet.Kartesianset.xmin;
        If (IntMax>DrawSet.Kartesianset.xmax) then IntMax:=DrawSet.Kartesianset.xmax;
        xMinThread:=IntMin+(TNumber*(IntMax-IntMin)/DrawSet.PerformanceSet.ThreadCount);
        xMaxThread:=xMinThread+((IntMax-IntMin)/DrawSet.PerformanceSet.ThreadCount);
        xMinThread_Canv:=round(XrealToCanv(xminThread));       //Für den ersten Thread könnte hier halbe Strichstärke (Knubbeldicke) abgezogen werden, falls noch im Screenbereich.
        ThreadResX:=trunc((XrealToCanv(IntMax)-XRealToCanv(IntMin))/DrawSet.PerformanceSet.ThreadCount)+1;
        // Wenn der letzte Thread Pixel anlegt, die außerhalb der Auflösung liegen, gibts Stress mit der LazIntfImage
        If (TNumber=DrawSet.PerformanceSet.ThreadCount-1) then begin //IntMax kann nicht größer xmax sein, da vorher genau in diesem Fall kastriert
          //ThreadresX sollte im ersten und letzten Thread um halbe Stichbreite oder halbe Knubbeldicke erhöht werden, falls noch im Screen darstellbar
          If DrawSet.PerformanceSet.AA_Mode>2 then ThreadResX:=ThreadResX+trunc(DrawSet.PerformanceSet.CurvWidth/2)+1;
          If xMinThread_Canv+ThreadResX>DrawSet.MainSizing.DrawResX then ThreadResX:=DrawSet.MainSizing.DrawResX-xMinThread_Canv;
        end;
        If (TNumber=0) then begin
          If (DrawSet.PerformanceSet.AA_Mode=3) or (DrawSet.PerformanceSet.AA_Mode=4) then begin
            ThreadResX:=ThreadResX+trunc(DrawSet.PerformanceSet.CurvWidth/2)+1;
            xMinThread_Canv:=xMinThread_Canv-(trunc(DrawSet.PerformanceSet.CurvWidth/2)+1);
          end;
          If xMinThread_Canv<0 then begin
            ThreadResX:=ThreadResX+xMinThread_Canv;
            xMinThread_Canv:=0;
          end;
        end;
        setlength(PixArray,0);
        setlength(PixPlan,0,0);
        setlength(AnteilPlan,0);
        Setlength(subpixels,0);
        Setlength(Qsubpixels,0);
        setlength(PCloud,0);
        IndividualOffset:=0;
        ResetPixMap;
        ThreadIsValid:=true;
        ana:=nil;
      end else ThreadIsValid:=false;
   end;

//=====================================================================================================================

  constructor tGraphCom.create(CreateSuspended:boolean; props:tGlobalSet; TNumber:byte; IntMin_im,IntMax_im:double; const PointCloud:tPointCloud);
    begin
      inherited create(CreateSuspended);
      sharedCreate(props,TNumber,IntMin_im,IntMax_im);
      PCloud:=PointCloud;
      funktion:='';
      If length(PCloud)=0 then threadIsValid:=false;
      GraphCompMode:=2;
    end;

//=====================================================================================================================

  constructor tGraphCom.create(CreateSuspended:boolean; props:tGlobalSet; TNumber:byte; IntMin_im,IntMax_im:double; funcString:string);
   var
     Konsole_GR:tKonsole;
     t:double;
    begin
      t:=time;
      inherited create(CreateSuspended);
      sharedCreate(props,TNumber,IntMin_im,IntMax_im);
      Konsole_GR:=tKonsole.create;                       //Ist das wirklich notwendig? Wird die Funktion nicht schon in der DrawClass zu genüge vereinfacht?
      Funktion:=Konsole_GR.BIOS(funcString);
      Konsole_GR.free;
      GraphCompMode:=0;
      If Funktion='' then ThreadIsValid:=false;
      stats.parser:=round((time-t)*86400000);
    end;

//=====================================================================================================================

  Destructor tGraphCom.free;
      begin
        PixArray:=nil;
        PixPlan:=nil;
        Subpixels:=nil;
        QSubPixels:=nil;
        AnteilPlan:=nil;
        inherited destroy;
      end;

//=====================================================================================================================

  procedure tGraphCom.SetIntegralOffset(offset_imp:double);
    begin
      IndividualOffset:=Offset_imp;
      GraphCompMode:=1;
    end;

//=====================================================================================================================

  function tGraphCom.GetBit(const Value: DWord; const Bit: Byte): Boolean;
    begin
      Result := (Value and (1 shl Bit)) <> 0;
    end;

//=====================================================================================================================

  function tGraphCom.SetBit(const Value: DWord; const Bit: Byte): DWord;
    begin
      Result := Value or (1 shl Bit);
    end;

//=====================================================================================================================

  function tGraphCom.QGetBit(const Value: QWord; const Bit: Byte): Boolean;
    begin
      Result := (Value and (Qword(1) shl Bit)) <> 0;
    end;

//=====================================================================================================================

  function tGraphCom.QSetBit(const Value: QWord; const Bit: Byte): QWord;
    begin
      Result := Value or (QWord(1) shl Bit);
    end;

//=====================================================================================================================

  procedure tGraphCom.ResetPixMap;
    var
     xres,yres,i,n:word;
     start,ende:extended;
     begin
      start:=Time;
      xres:=ThreadResX;
      yres:=DrawSet.MainSizing.DrawResY;
      setlength(PixPlan,xres,yres);
      for i:=0 to xres-1 do begin
        for n:=0 to yres-1 do begin
          PixPlan[i,n]:=2147483647;
        end;
      end;
      Ende:=Time;
      stats.Resettime:=round((ende-start)*24*60*60*1000);
    end;

//=====================================================================================================================

  procedure tGraphcom.generate_Results;
     var
       laenge,i,x,y,yres_loc,xIntMin,xAxis,yAxis:longint;
       bgrdCol,bgrdCol2,CurvCol,CurvCol2,bgrd_Col_act,Curv_Col_Act:tColor;
       bgrdGradr,BgrdGradg,BgrdGradb,CurvGradr,CurvGradb,CurvGradg,gen_buf,gen_buf_rest,xAkt,yAkt,xmin,xmax,ymin,ymax,xValueCor,yValueCor,ReziGDichteX,ReziGDichteY,ReziAA_total_loc:double;
       start,ende:extended;
       gbx,gby:array [0..1]of longint;
      begin
        start:=time;

        xValueCor:=1; yValueCor:=1;
        If DrawSet.Kartesianset.UseUsrDefUnits then begin
          If DrawSet.Kartesianset.UnitSignX<>'' then xValueCor:=DrawSet.Kartesianset.UnitValueX;
          If DrawSet.Kartesianset.UnitSignY<>'' then yValueCor:=DrawSet.Kartesianset.UnitValueY;
        end;

        xmin:=DrawSet.Kartesianset.xmin; xmax:=DrawSet.Kartesianset.xmax;
        ymax:=DrawSet.Kartesianset.ymax; ymin:=DrawSet.Kartesianset.ymin;
        xAxis:=round(yRealToCanv(0));
        yAxis:=round(xRealToCanv(0));
        ReziGDichteX:=1/(DrawSet.Kartesianset.GridDensX/xValueCor);
        ReziGDichteY:=1/(DrawSet.Kartesianset.GridDensY/yValueCor);
        If 1/ReziGDichteX*(xMax-xMin)>DrawSet.MainSizing.DrawResX then ReziGdichteX:=1/(DrawSet.MainSizing.DrawResX/(xMax-xMin));
        If 1/ReziGDichteY*(yMax-yMin)>DrawSet.MainSizing.DrawResY then ReziGdichtey:=1/(DrawSet.MainSizing.DrawResY/(yMax-yMin));
        bgrdCol:=DrawSet.ColorSet.bgrdCol;
        bgrdCol2:=DrawSet.ColorSet.bgrdCol2;
        CurvCol:=DrawSet.ColorSet.CurvCol;
        CurvCol2:=DrawSet.ColorSet.CurvCol2;
        xIntMin:=round(xRealToCanv(IntMin));
        ReziAA_total_loc:=1/(DrawSet.PerformanceSet.AA_Level_hor*DrawSet.PerformanceSet.AA_Level_ver);
        yres_loc:=DrawSet.MainSizing.DrawResY;
        Case DrawSet.ColorSet.bgrdGrad of      //Hintergrund-Gradient bestimmen
          0: begin     //vertikal
             BgrdGradr:=(red(bgrdCol2)-red(bgrdCol))/yRes_loc;
             BgrdGradg:=(green(bgrdCol2)-green(bgrdCol))/yres_loc;
             BgrdGradb:=(blue(bgrdCol2)-blue(bgrdCol))/yRes_loc;
          end;
          1: begin     //horizontal
             BgrdGradr:=(red(bgrdCol2)-red(bgrdCol))/DrawSet.MainSizing.DrawResX;
             BgrdGradg:=(green(bgrdCol2)-green(bgrdCol))/DrawSet.MainSizing.DrawResX;
             BgrdGradb:=(blue(bgrdCol2)-blue(bgrdCol))/DrawSet.MainSizing.DrawResX;
          end;
        end;
        Case DrawSet.ColorSet.CurvGrad of     //Kurvengradient bestimmen
          0: begin      //vertikal
             CurvGradr:=(red(CurvCol2)-red(CurvCol))/yRes_loc;
             CurvGradg:=(green(CurvCol2)-green(CurvCol))/yres_loc;
             CurvGradb:=(blue(CurvCol2)-blue(CurvCol))/yRes_loc;
          end;
          1: begin      //horizontal
             gen_buf:=(IntMax-IntMin)/(DrawSet.Kartesianset.xmax-DrawSet.Kartesianset.xmin)*DrawSet.MainSizing.DrawResX;
             CurvGradr:=(red(CurvCol2)-red(CurvCol))/gen_buf;
             CurvGradg:=(green(CurvCol2)-green(CurvCol))/gen_Buf;
             CurvGradb:=(blue(CurvCol2)-blue(CurvCol))/gen_Buf;
          end;
        end;
        laenge:=length(PixArray)-1;
        //-----> Gitternetz-Berücksichtigung start
        xAkt:=(trunc(1/ReziGDichteX*xmin)-1)*ReziGDichteX;
        yAkt:=(trunc(1/ReziGDichteY*Ymax)+1)*ReziGDichteY;
        gby[0]:=round(yRealToCanv(yAkt));
        i:=trunc((PixArray[0].Punkt.y-gby[0])*1/ReziGDichteY/zoomy);
        yAkt:=(trunc(1/ReziGDichteY*Ymax)+1)*ReziGDichteY-i*ReziGDichteY;
        for i:=0 to 1 do begin
          gbx[i]:=round(xRealToCanv(xAkt));
          gby[i]:=round(yRealToCanv(yAkt));
          if i=0 then xAkt:=xAkt+(ReziGdichteX);
          if i=0 then yAkt:=yAkt-(ReziGdichteY);
        end;
        //-----> Gitternetz-Berücksichtigung ende
        for i:=0 to laenge do begin
          x:=PixArray[i].Punkt.x;
          y:=PixArray[i].Punkt.y;
          //----> Gitternetzberücksichtigung start
          if (x>=gbx[1]) then begin     //Berücksichtigung vertikaler Striche, suche neuer Werte
            gbx[0]:=gbx[1];
            xAkt:=xAkt+(ReziGdichteX);
            gbx[1]:=round(xRealToCanv(xAkt));
          end;
          if (y<gby[0]) or (y>gby[1]) then begin   //Berücksichtigung horizontaler Striche, suche neuer Werte
            if y<gby[0] then begin
              gby[1]:=gby[0];
              If gby[0]<>round(yRealToCanv(yakt))then yAkt:=yAkt+(2*ReziGdichteY) else yAkt:=yAkt+(ReziGdichteY);
              gby[0]:=round(yRealToCanv(yAkt));
            end else begin
              gby[0]:=gby[1];
              If gby[1]<>round(yRealToCanv(yakt))then yAkt:=yAkt-(2*ReziGdichteY) else yAkt:=yAkt-(ReziGdichteY);
              gby[1]:=round(yRealToCanv(yAkt));
            end;
          end;
          //----> Gitternetzberücksichtigung ende
          //Berechnung der Hintergrundfarbe in Abhängigkeit eines Netz- oder Achsentreffers
          if ((x=gbx[0])or(x=gbx[1])or(y=gby[0])or(y=gby[1]))  and  (x<>yAxis)and(y<>xAxis) then bgrd_Col_Act:=DrawSet.ColorSet.GridCol  //Wenn x oder y= Netz, aber nicht Achse
          else if (abs(y-xAxis)<=1) or (abs(x-yAxis)<=1)then begin    //Wenn x oder y = Achse
            If (y=xAxis) or (x=yAxis) then bgrd_Col_act:=DrawSet.ColorSet.AxisCol //Wenn Kern der x-Achse oder y-Achse
            else begin           //Wenn Rand der einen oder der anderen Achse
              if (red(DrawSet.ColorSet.AxisCol)+120)<256 then bgrd_Col_Act:=red(DrawSet.ColorSet.AxisCol)+120 else bgrd_Col_Act:=255;
              if (green(DrawSet.ColorSet.AxisCol)+120)<256 then bgrd_Col_Act:=bgrd_Col_Act*256+green(DrawSet.ColorSet.AxisCol)+120 else bgrd_Col_Act:=bgrd_Col_Act*256+255;
              if (blue(DrawSet.ColorSet.AxisCol)+120)<256 then bgrd_Col_Act:=bgrd_Col_Act*256+blue(DrawSet.ColorSet.AxisCol)+120 else bgrd_Col_Act:=bgrd_Col_Act*256+255;
            end;
          end else begin //Wenn weder Netz noch eine der beiden Achsen getroffen
            Case DrawSet.ColorSet.bgrdGrad of
              -1:bgrd_Col_act:=bgrdCol;
              0:bgrd_Col_act:=rgbToColor(trunc(BgrdGradr*y+red(bgrdCol)),trunc(BgrdGradg*y+green(bgrdCol)),trunc(BgrdGradb*y+blue(bgrdCol)));
              1:bgrd_Col_act:=rgbToColor(trunc(BgrdGradr*x+red(bgrdCol)),trunc(BgrdGradg*x+green(bgrdCol)),trunc(BgrdGradb*x+blue(bgrdCol)));
            end;
          end;
          //Berechnung der Kurvenfarbe
          Case DrawSet.ColorSet.CurvGrad of
            -1:Curv_Col_act:=CurvCol;
            0:Curv_Col_act:=rgbToColor(trunc(CurvGradr*y+red(CurvCol)),trunc(CurvGradg*y+green(CurvCol)),trunc(CurvGradb*y+blue(CurvCol)));
            1:Curv_Col_act:=rgbToColor(trunc(CurvGradr*(x-xIntMin)+red(CurvCol)),trunc(CurvGradg*(x-xIntMin)+green(CurvCol)),trunc(CurvGradb*(x-xIntMin)+blue(CurvCol)));
          end;
          gen_buf:=AnteilPlan[i]*ReziAA_total_loc;
          gen_buf_rest:=1-gen_buf;   //bringt keine Vorteile!
          //Berechnung der gemischten, endgültigen Pixel-Farbe
          PixArray[i].Farbe:=RGBToColor(trunc((red(Curv_Col_act)*gen_buf)+(red(bgrd_Col_act)*(gen_buf_rest))),
                                        trunc((green(Curv_Col_act)*gen_buf)+(green(bgrd_Col_act)*(gen_buf_rest))),
                                        trunc((blue(Curv_Col_act)*gen_buf)+(blue(bgrd_Col_act)*(gen_buf_rest))));




        end;
        Ende:=time;
        stats.postprocess:=round((ende-start)*24*60*60*1000);

      end;

//=====================================================================================================================

  function tGraphCom.GetPixArray:tPixArray;
    begin
      GetPixArray:=PixArray;
    end;

//=====================================================================================================================

  function tGraphCom.GetTimes:tThreadstats;
    begin
      result:=stats;
    end;

//=====================================================================================================================

  function tGraphCom.XRealToCanv (const x:double):double;              //Ungerundeter Wert!!!
    begin
      result:=(x-DrawSet.Kartesianset.Xmin)*zoomX;
    end;

//=====================================================================================================================

  function tGraphCom.YRealToCanv (const Y:double):double;              //Ungerundeter Wert!!!
   begin
     result:=(DrawSet.Kartesianset.ymax-Y)*zoomY;
   end;

//=====================================================================================================================

  procedure tGraphCom.C_unbufferedTetragonAACurv(const ComputationMode:byte);
   var
     Start,Ende:extended;
     mw1,mw2,beta,mh,mt:double;//Steigung der Winkelhalbierenden und der oberen und unteren Begrenzung
     p1h,p1t,p2h,p2t:tVec2D;
     m:array [1..2]of double;
     p: array [1..3]of tVec2D;
     Error:array [1..3]of integer;
     ScanArea:trect;
     i,z,k,n,Anteil,index,lPixArr,yres_loc,DrawLimit,xres_loc,AA_hor_loc,AA_ver_loc,PixArrMgr,AATotal,CloudIndex,CloudSize,xmin_arr:longint;
     loc_StepWidth,IntValue,xfunc,xMax_loc,yBuf,invAA_ver,invAA_hor,AA_ver_startp,AA_hor_startp,h_CurvWidth,subpix_x,subpix_y:double;
     CoverRun:integer=-1;
     pitchSaved:boolean=false;
   const
     PixArrMgrJump=128;    //PixArray wird alle PixArrMgrJump Schritte um PixArrMgrJump+1 vergrößert anstatt bei jedem Schritt um ein Feldelement
    begin
      Start:=time;
      stats.StepCount:=0;
    //---------------------------- tAnalyse erzeugen und funktion zuweisen

        If (ComputationMode=2)or((ana<>nil)and(Ana.Error=0)) then begin
          //---->> Analyse-Bedingungen setzen
          //--->> Zunächst für alle Computationmodes geltend
          PixArrMgr:=PixArrMgrJump+1; //Wenn PixArrMgr größer wird als PixArrMgrJump, dann wird das Array um PixArrMgrJump vergrößert (große Zeitvorteile)
          lPixArr:=length(PixArray);
          AA_hor_loc:=DrawSet.PerformanceSet.AA_Level_hor;
          AA_ver_loc:=DrawSet.PerformanceSet.AA_Level_ver;
          AATotal:=AA_hor_loc*AA_ver_loc;
          h_CurvWidth:=DrawSet.PerformanceSet.CurvWidth/2;
          xmin_arr:=xMinThread_Canv;
          xres_loc:=ThreadResX;
          yres_loc:=DrawSet.MainSizing.DrawResY;
          DrawLimit:=yres_loc+800;
          CloudSize:=length(PCloud);
          CloudIndex:=-1;
          //--> Oft gebrauchte, über die gesamte Thread-Dauer konstante bleibende Werte ausrechnen, ums nicht 10^7 mal berechnen zu müssen
          invAA_ver:=1/AA_ver_loc;
          invAA_hor:=1/AA_hor_loc;
          AA_ver_startp:=-0.5+0.5/AA_ver_loc;
          AA_hor_startp:=-0.5+0.5/AA_hor_loc;


          xMax_loc:=xMaxThread+(h_Curvwidth/zoomx*3);
          //die Drei ersten Punkte berechnen zur Bestimmung der Winkelhalbierenden zwischen linie 1 (p1p2) und linie 2 (p2p3) und Winkelhalbierendensteigung mw2===
          If ComputationMode<>2 then begin
            if DrawSet.PerformanceSet.unsharpnes<>0 then loc_StepWidth:=DrawSet.PerformanceSet.unsharpnes/zoomx else loc_StepWidth:=0.025/zoomx;
            //---> Um sichtbare Übergänge zwischen den Threads zu entfernen, muss man min. eine Strichbeite am Anfang und Ende zu viel Berechnen pro Thread
            If ComputationMode=0 then begin
              xfunc:=xMinThread-(h_Curvwidth/zoomx*3); //Um sichtbare Übergänge zwischen den Threads zu entfernen, muss man min. 1,5 Strichbeiten am Anfang und Ende zu viel Berechnen pro Thread
              For i:=1 to 3 do begin
                p[i].x:=xRealToCanv(xfunc);
                ybuf:=yRealToCanv(Ana.FWert(xFunc));
                if ybuf>DrawLimit then ybuf:=DrawLimit;
                if ybuf<-800 then ybuf:=-800;
                p[i].y:=ybuf;
                Error[i]:=Ana.Error;
                If (i>1)and(drawset.PerformanceSet.unsharpnes=0) then begin
                  m[i-1]:=(ybuf-p[i-1].y)/(p[i].x-p[i-1].x);
                  {if i=2 then begin
                    if abs(m[i-1])>1 then loc_Stepwidth:=1/abs(zoomx*m[i-1]);
                  end;
                  if i=3 then begin
                    if abs(m[2])>=abs(m[1]) then loc_Stepwidth:=1/abs(zoomx*m[2]) else PitchSaved:=true;
                  end;}
                end;
                xFunc:=xFunc+loc_StepWidth;
              end;
            end else begin   //1,5 Strichstärken vor und nach dem Thread führen bei hochfreq. Funktionen zu besseren Erg.
              xFunc:=xMinThread;
              IntValue:=IndividualOffset;
              For i:=1 to 3 do begin
                p[i].x:=xRealToCanv(xfunc);
                ybuf:=yRealToCanv(IndividualOffset);
                if ybuf>DrawLimit then ybuf:=DrawLimit;
                if ybuf<-800 then ybuf:=-800;
                p[i].y:=ybuf;
                Error[i]:=Ana.Error;
                xFunc:=xFunc+loc_StepWidth;
              end;
            end;
          end else begin
            CloudIndex:=3;
            Error[1]:=0;Error[2]:=0;Error[3]:=0;     //Wenn Error <>0 dann, werden keine Ausgaben für einen Punkt berechnet. Error braucht man nur für richtige Funktionen
            for i:=1 to 3 do begin
              p[i].x:=xRealToCanv(PCloud[i-1].x);
              P[i].y:=yRealToCanv(PCloud[i-1].y);
            end;
            ybuf:=0;
            xFunc:=xMinThread;
          end;
          beta:=((arctan((p[2].y-p[1].y)/(p[2].x-p[1].x))+arctan((p[3].y-p[2].y)/(p[3].x-p[2].x)))/2)+pi/2;  //Winkel der Winkelhalbierenden
          mw2:=tan(beta); //Steigung der Winkelhalbierenden zwischen linie 1 (p1p2) und linie 2 (p2p3)
          p2h.x:=p[2].x+h_CurvWidth*cos(beta); p2h.y:=p[2].y+h_CurvWidth*sin(beta);
          p2t.x:=p[2].x-h_CurvWidth*cos(beta); p2t.y:=p[2].y-h_CurvWidth*sin(beta);

      //Hauptwiederholschleife=======================================================================================================================================
      Repeat
      //p1 und p2 holen
        p[1]:=p[2]; p[2]:=p[3];
        p1h:=p2h; p1t:=p2t; mw1:=mw2;
        Error[1]:=Error[2];Error[2]:=Error[3];
        Case ComputationMode of
          0: begin
            p[3].x:=xRealToCanv(xfunc);
            ybuf:=yRealToCanv(Ana.FWert(xFunc));
            Error[3]:=ana.Error;
            if ybuf>DrawLimit then ybuf:=DrawLimit;
            if ybuf<-800 then ybuf:=-800;
          end;
          1: begin
            p[3].x:=xRealToCanv(xfunc);
            IntValue:=IntValue+(Ana.FWert(xFunc+loc_stepwidth*0.5)*loc_Stepwidth);
            ybuf:=yRealToCanv(intValue);
            Error[3]:=ana.Error;
            if ybuf>DrawLimit then ybuf:=DrawLimit;
            if ybuf<-800 then ybuf:=-800;
          end;
          2: begin
            p[3].x:=xRealToCanv(PCloud[CloudIndex].x);
            ybuf:=yRealToCanv(PCloud[CloudIndex].y);
          end;
        end;

        p[3].y:=ybuf;
        beta:=((arctan((p[2].y-p[1].y)/(p[2].x-p[1].x))+arctan((p[3].y-p[2].y)/(p[3].x-p[2].x)))*0.5)+pi*0.5;  //Winkelhalbierende
        If (abs(beta)<>Pi*0.5) then mw2:=tan(beta)
          else mw2:=sign(beta)*20001;
        p2h.x:=p[2].x+h_CurvWidth*cos(beta);       //prefetch von h_CurvWidth
        p2h.y:=p[2].y+h_CurvWidth*sin(beta);
        p2t.x:=p[2].x-h_CurvWidth*cos(beta);
        p2t.y:=p[2].y-h_CurvWidth*sin(beta);
        mh:=(p2h.y-p1h.y)/(p2h.x-p1h.x);
        mt:=(p2t.y-p1t.y)/(p2t.x-p1t.x);
        If (((p[1].y<yres_loc+40)and(p[1].y>-41))or((p[2].y<yres_loc+40)and(p[2].y>-41)))and((error[1]=0) and (Error[2]=0)) then begin

          If p1h.x<p1t.x then ScanArea.Left:=round(p1h.x) else ScanArea.Left:=round(p1t.x);      //Betroffene Pixel in x-Richtung abstecken
          If p2h.x>p2t.x then ScanArea.right:=round(p2h.x) else ScanArea.right:=round(p2t.x);    //Betroffene Pixel in x-Richtung abstecken
          If p1h.y>p2h.y then Scanarea.Top:=round(p1h.y) else Scanarea.top:=round(p2h.y);       //Betroffene Pixel in y-Richtung abstecken
          If p1t.y<p2t.y then Scanarea.Bottom:=round(p1t.y) else Scanarea.Bottom:=round(p2t.y);       //Betroffene Pixel in y-Richtung abstecken

          for i:=ScanArea.Left to ScanArea.Right do begin //i entspricht x-Richtung für Pixel
            for z:=ScanArea.bottom to ScanArea.Top do begin  //z entspricht y-Richtung für Pixel
              Anteil:=0;
              If h_CurvWidth>1.45 then begin
                If ((cos(abs(arctan(mh))))*(z-(p1h.y+mh*(i-p1h.x)))<-0.75) and ((cos(abs(arctan(mt))))*(z-(p1t.y+mt*(i-p1t.x)))>0.75) then begin
                  Anteil:=AATotal;
                  If (i-xMin_Arr>-1)and (i-xMin_Arr<xRes_loc) and(z>-1)and(z<yres_loc) then begin
                  Index:=PixPlan[i-xMin_Arr,z];
                  If Index=2147483647 then begin  //Wenn Pixel noch nicht vorhanden
                    If PixArrMgr>PixArrMgrJump then begin
                      Setlength(PixArray,lPixArr+PixArrMgrJump+1);
                      Setlength(Anteilplan,lPixArr+PixArrMgrJump+1);
                      PixArrMgr:=0;
                    end;
                    //Setlength(PixArray,lPixArr+1);
                    PixArray[lPixArr].Punkt.x:=i;
                    PixArray[lPixArr].Punkt.y:=z;      //trunc statt round bringt keinen P-Vorteil, jedoch ist das Ergebnis sicher unter 255 pro kanal
                    AnteilPlan[lPixArr]:=Anteil;
                    PixPlan[i-xMin_Arr,z]:=lPixArr;
                    inc(lPixArr,1);
                    inc(PixArrMgr,1);
                  end else begin  //Wenn Pixel bereits vorhanden. Anteil der Hintergrundfarbe muss rausgerechnet werden für jeden dazugekommenen Kurvenanteil
                    AnteilPlan[index]:=AATotal;
                  end;
                end;
                end;
              end;
              If Anteil=0 then begin
              For k:=0 to (AA_hor_loc-1) do begin   //Subpixel durchgehen
                for n:=0 to (AA_ver_loc-1) do begin
                  Subpix_x:=i+AA_hor_startp+k*invAA_hor;   //Würde ein Prefetch von AA_hor_loc und AA_ver_loc was tempo-mäßiges bringen?
                  Subpix_y:=z+AA_ver_startp+n*invAA_ver;
                  If (CoverRun=0) then begin
                    If (Subpix_y<p1h.y+mh*(subpix_x-p1h.x))and(Subpix_y>p1t.y+mt*(subpix_x-p1t.x)) then // Wenn tiefer als hochlinie und höher als tieflinie
                      If ((mw1>0)and(((subpix_x-p1t.x)*mw1+p1t.y)>subpix_y))  or  ((mw1<0)and(((subpix_x-p1t.x)*mw1+p1t.y)<subpix_y))then //Wenn rechts von der ersten Winkelhalbierenden
                        if ((mw2>0)and(((subpix_x-p2t.x)*mw2+p2t.y)<subpix_y))  or ((mw2<0)and(((subpix_x-p2t.x)*mw2+p2t.y)>subpix_y))//Wenn links von der zweiten (aktuellen) Winkelhalbierenden
                          then inc(Anteil,1);

                  end else begin
                    If (Subpix_y<p1h.y+mh*(subpix_x-p1h.x))and(Subpix_y>p1t.y+mt*(subpix_x-p1t.x)) then
                        if ((mw2>0)and(((subpix_x-p2t.x)*mw2+p2t.y)<subpix_y))  or ((mw2<0)and(((subpix_x-p2t.x)*mw2+p2t.y)>subpix_y))
                          then inc(Anteil,1);
                  end;
                end;//AA_ver_loc
              end;//AA_hor_loc
              If Anteil>0 then begin     //mindestens 1 Subpixel von Kurve getroffen
                If (i-xMin_Arr>-1)and (i-xMin_Arr<xRes_loc) and(z>-1)and(z<yres_loc) then begin
                  Index:=PixPlan[i-xMin_Arr,z];
                  If Index=2147483647 then begin  //Wenn Pixel noch nicht vorhanden
                    If PixArrMgr>PixArrMgrJump then begin
                      Setlength(PixArray,lPixArr+PixArrMgrJump+1);
                      Setlength(AnteilPlan,lPixArr+PixArrMgrJump+1);
                      PixArrMgr:=0;
                    end;
                    //Setlength(PixArray,lPixArr+1);
                    PixArray[lPixArr].Punkt.x:=i;
                    PixArray[lPixArr].Punkt.y:=z;      //trunc statt round bringt keinen P-Vorteil, jedoch ist das Ergebnis sicher unter 255 pro kanal
                    AnteilPlan[lPixArr]:=Anteil;
                    PixPlan[i-xMin_Arr,z]:=lPixArr;
                    inc(lPixArr,1);
                    inc(PixArrMgr,1);
                  end else begin  //Wenn Pixel bereits vorhanden. Anteil der Hintergrundfarbe muss rausgerechnet werden für jeden dazugekommenen Kurvenanteil
                    If Anteil+AnteilPlan[index]<=AATotal then AnteilPlan[index]:=Anteil+AnteilPlan[index] else AnteilPlan[index]:=AATotal;
                  end;
                end;
              end; //If Anteil >0
            end;
            end;
          end;//endfor ScanArea
        end;//end Checks
        If drawset.PerformanceSet.unsharpnes=0 then begin
          m[1]:=m[2];
          m[2]:=(p[3].y-p[2].y)/(p[3].x-p[2].x);
          if (abs(m[2])>=abs(m[1]))then begin
            if abs(m[2])>10000 then m[2]:=sign(m[2])*10000;
            if (abs(m[2])>1)then loc_Stepwidth:=2/abs(zoomx*m[2]) else loc_Stepwidth:=1/zoomx;
          end else begin
            if not PitchSaved then pitchSaved:=true else begin
              PitchSaved:=false;
              if abs(m[1])>1 then loc_Stepwidth:=1/abs(zoomx*m[1]) else loc_Stepwidth:=1/zoomx;;
            end;
          end;
        end;
        xFunc:=xFunc+loc_StepWidth;
        if CoverRun=-1 then CoverRun:=0 else if  ((xFunc+loc_StepWidth>=xmax_loc)or((ComputationMode=2)and(CloudIndex>=CloudSize))) then CoverRun:=1;
        inc(stats.StepCount,1);
      until (xFunc>=xmax_loc)or(CloudIndex>=CloudSize); //endwhile xFunc<xmax
      if lPixArr>0 then begin
        setlength(PixArray,lPixArr);
        setlength(AnteilPlan,lPixArr);
      end;
      if length(PixArray)>0 then generate_results;
      end;//if ana.error=0 begin

      Ende:=time;
      stats.RenderTime:=round((ende-start)*24*60*60*1000)-stats.postprocess;
    end;

//=====================================================================================================================

  procedure tGraphCom.C_BufferedHexagonAACurv(const ComputationMode:byte);
   var
     Start,Ende:extended;
     p1,p2,p1h,p1t,p2h,p2t,p1p2,subpix:tVec2D;
     ScanArea:trect;
     Error1,Error2,index,lPixArr,yres_loc,xres_loc,Anteil,i,z,intbuf,xmin_arr,cloudIndex,CloudSize:longInt;
     loc_StepWidth,xfunc,xMax_loc,yBuf,IntValue,invAA_ver,invAA_hor,AA_ver_startp,AA_hor_startp,h_CurvWidth,genbuf,m,beta:double;
     hit:boolean;
     DSubPixBuf,DrawLimit:DWord;
     QSubPixBuf:Qword;
     ColumnStartIndex,k,n,AA_hor_loc,AA_ver_loc,AA_Total,PixArrMgr:byte;
   const
     PixArrMgrJump=128;    //PixArray wird alle PixArrMgrJump Schritte um PixArrMgrJump+1 vergrößert anstatt bei jedem Schritt um ein Feldelement, max. 254
    begin
     Start:=time;
      //---->> tAnalyse erzeugen und funktion zuweisen

        //---->> Berechnungsbedingungen setzen und abhängig vom Berechnungsmodus Graphberechnung starten
        If (ComputationMode=2)or((ana<>nil)and(Ana.Error=0)) then begin
          //--->> Zunächst für alle Computationmodes geltend
          PixArrMgr:=PixArrMgrJump+1; //Wenn PixArrMgr größer wird als PixArrMgrJump, dann wird das Array um PixArrMgrJump vergrößert (große Zeitvorteile)
          lPixArr:=length(PixArray);  //sollte folglich 0 sein, wenn erste funktion für die Klasse
          //--> Anti-Aliasing-Optionen setzen
          AA_hor_loc:=DrawSet.PerformanceSet.AA_Level_hor; AA_ver_loc:=DrawSet.PerformanceSet.AA_Level_ver; AA_Total:=AA_hor_loc*AA_ver_loc;
          h_CurvWidth:=DrawSet.PerformanceSet.CurvWidth/2; //halbe Kurvenbreite
          xmin_arr:=xMinThread_Canv;                       //Thread-bezogener Array-Offset für Canvas-x-Werte
          xres_loc:=ThreadResX;                            //Thread-bezogene Pixelauflösung in X
          yres_loc:=DrawSet.MainSizing.DrawResY;           //y-Res ist für alle Threads=Bildhöhe, da threads auf Ordinate sektioniert
          DrawLimit:=yres_loc+1500;                        //Wenn ein Wert 1500 Pixel unterhalb der Zeichenfläche ist. Alles noch tiefer wird ignoriert und auf DrawLimit gesetzt
          CloudSize:=length(PCloud);                       //Größe für Punktewolke (falls vorhanden) ermitteln
          CloudIndex:=-1;                                  //Fortschritts-Index innerhalb der Punktewolke zurücksetzen
          xMax_loc:=xMaxThread+(h_Curvwidth/zoomx*3);

          //--> Oft gebrauchte, über die gesamte Thread-Dauer konstante bleibende Werte ausrechnen, ums nicht 10^7 mal berechnen zu müssen
          invAA_ver:=1/AA_ver_loc;
          invAA_hor:=1/AA_hor_loc;
          AA_ver_startp:=-0.5+0.5*invAA_ver;  //Vertikaler Subpixeloffset oben
          AA_hor_startp:=-0.5+0.5*InvAA_hor;  //horizontaler Subpixeloffset links
          //--> Wenn kein Punktewolkenmodus
          If ComputationMode<>2 then begin
            //-> Startpunkte berechnen
            loc_StepWidth:=DrawSet.PerformanceSet.unsharpnes/zoomx;
            If ComputationMode=0 then begin
              //-> Zeichenvorgang einer "normalen" Funktion
              xfunc:=xMinThread-(h_Curvwidth/zoomx*3); //Um sichtbare Übergänge zwischen den Threads zu entfernen, muss man min. 1,5 Strichbeiten am Anfang und Ende zu viel Berechnen pro Thread
               //Um sichtbare Übergänge zwischen den Threads zu entfernen, muss man min. 1,5 Strichbeiten am Anfang und Ende zu viel Berechnen pro Thread
              p1.x:=xRealToCanv(xfunc);
              ybuf:=yRealToCanv(Ana.FWert(xFunc));
            end else begin
              //-> Zeichenvorgang einer "normalen" Funktion
              xFunc:=xMinThread; //im Modus für numerische Integrale darf man nicht zurücksetzen.
              p1.x:=xRealToCanv(xfunc);
              IntValue:=IndividualOffset;
              ybuf:=yRealToCanv(IndividualOffset);
            end;
            //---> yBuf in Canvas-Koordinaten macht keinen sichtbaren Unterschied, ob -10E15 oder  -1500
            if ybuf>DrawLimit then ybuf:=DrawLimit;
            if ybuf<-1500 then ybuf:=-1500;
            p1.y:=ybuf;
            Error1:=Ana.Error;
            xFunc:=xFunc+loc_StepWidth;
          end else begin //-> Wenn Punktewolken-Modus
            CloudIndex:=1;
            error1:=0;error2:=0;//Wenn Error <>0 dann, werden keine Ausgaben für einen Punkt berechnet. Error braucht man nur für richtige Funktionen
            p1.x:=xRealToCanv(PCloud[0].x);
            P1.y:=yRealToCanv(PCloud[0].y);
            ybuf:=0;
            xFunc:=xMinThread;
          end;
          //---->> Hauptwiederholschleife
          Repeat
            //---> Neuen P2 ausrechnen
            Case ComputationMode of
              0: begin                                 //normaler Berechnungsmodus
                p2.x:=xRealToCanv(xfunc);
                ybuf:=yRealToCanv(Ana.FWert(xFunc));
                Error2:=ana.Error;
              end;
              1: begin                                 //numerischer Integrationsmodus
                p2.x:=xRealToCanv(xfunc);
                IntValue:=IntValue+(Ana.FWert(xFunc+loc_stepwidth*0.5)*loc_Stepwidth);
                ybuf:=yRealToCanv(intValue);
                Error2:=ana.Error;
              end;
              2: begin                                 //Punktewolkenmodus
                p2.x:=xRealToCanv(PCloud[CloudIndex].x);
                ybuf:=yRealToCanv(PCloud[CloudIndex].y);
              end;
            end;
            if ybuf>DrawLimit then ybuf:=DrawLimit;
            if ybuf<-1500 then ybuf:=-1500;
            p2.y:=ybuf;
            //---> Wenn am Bildschirm überhaupt eine sichtbare Veränderung eintritt, dann muss der ganze AA-Algorithmus durchlaufen werden und anschließend wird der P1 der alte P2. Der neue P2 wird berechnet
            If (abs(p1.y-p2.y)>0.67)or(p2.x-p1.x>0.67)or(h_CurvWidth>1.95) then begin  //Bei dicken Kurven muss der Alg. immer durchlaufen werden, da sonst komische Ausbrüche bei Extremstellen hochfreq. Funktionen
              p1p2.x:=p2.x-p1.x; p1p2.y:=p2.y-p1.y;  //Richtungsvektor P1 nach P2
              m:=(p1p2.y)/(p1p2.x);
              beta:=arctan(m)+pi*0.5;
              //-> Rechteckecken bei P2 in Y
              genbuf:=sin(beta)*h_Curvwidth;
              p2h.y:=p2.y+genbuf;
              p2t.y:=p2.y-genbuf;
              //-> Rechteckecken bei P2 in X
              genbuf:=cos(beta)*h_Curvwidth;
              p2h.x:=p2.x+genbuf;       //prefetch von h_CurvWidth
              p2t.x:=p2.x-genbuf;
              //-> Rechteckecken bei P1 in X und Y
              p1h.x:=p2h.x-(p1p2.x); p1h.y:=p2h.y-(p1p2.y);
              p1t.x:=p2t.x-(p1p2.x); p1t.y:=p2t.y-(p1p2.y);
              //---> Wenn P1 oder P2 bis max. 40 Pix ausserhalb des sichtbaren Bereichs, werden noch Scanbereich erfasst, und evtl. Pixel, falls sichtbar, berechnet
              If (((p1.y<yres_loc+40)and(p1.y>-41))or((p2.y<yres_loc+40)and(p2.y>-41)))and((error1=0) and (Error2=0)) then begin
              //-> Scan-Rechteck festlegen
                If p1.y<p2.y then begin
                  ScanArea.Left:=round(p1h.x);    //Betroffene Pixel in x-Richtung abstecken
                  ScanArea.Right:=round(p2t.x);   //Betroffene Pixel in x-Richtung abstecken
                  ScanArea.Bottom:=round(p1t.y);  //Betroffene Pixel in y-Richtung abstecken
                  ScanArea.Top:=round(p2h.y);     //Betroffene Pixel in y-Richtung abstecken
                end else begin
                  ScanArea.left:=round(p1t.x);    //Betroffene Pixel in x-Richtung abstecken
                  ScanArea.Right:=round(p2h.x);   //Betroffene Pixel in x-Richtung abstecken
                  ScanArea.Bottom:=round(p2t.y);  //Betroffene Pixel in y-Richtung abstecken
                  ScanArea.Top:=round(p1h.y);     //Betroffene Pixel in y-Richtung abstecken
                end;
                genbuf:=1/sqrt(sqr(m)+1); //=cos(arctan(m)); Braucht man für Abstandsberechnungen der Subpixel zur Linie P1P2 (umrechnung delta_Y in orthogonal-Höhe)

                for i:=ScanArea.Left to ScanArea.Right do begin //i entspricht x-Richtung für Pixel, Pixelebene
                  for z:=ScanArea.bottom to ScanArea.Top do begin  //z entspricht y-Richtung für Pixel, Pixelebene
                    If (i-xMin_Arr>-1)and (i-xMin_Arr<xRes_loc) and(z>-1)and(z<yres_loc) then begin  //Wenn aktuelles Pixel im sichtbaren Bereich, wird auf niedrigerer Ebene weitergerechnet
                    hit:=false;
                    //-> Index für aktuelles Pixel im Ergebnis-Array wird aus der PixMap ausgelesen
                    Index:=PixPlan[i-xMin_Arr,z];
                    If Index <2147483647 then begin // Index leer=2147483647, <2147483647 heißt, der Pixel wurde schonmal angelegt an der durch die Nummer angezeigten Adresse
                      Anteil:=AnteilPlan[index];
                      intbuf:=anteil;    //intbuf wird verwendet um am Ende festzustellen, ob sich an einem Pixel der Anteil verändert hat
                      if Intbuf<AA_Total then If AA_Total<33 then DSubPixBuf:=Subpixels[index] else QSubPixBuf:=QSubpixels[index];
                    end else begin                  // Pixel wird mit Anteil=0 neu angelegt, falls Index=high(integer)
                      Anteil:=0;
                      intbuf:=0;
                      If AA_Total<33 then DSubPixBuf:=$0 else QSubPixBuf:=Qword($0);
                    end;
                    //-> kommt bei dicken Strichen zum Einsatz, wenn Pixel nicht bereits vollen Farbanteil hat
                    if (h_curvwidth>0.9)and(anteil<AA_Total) then begin
                      Subpix.x:=i;
                      Subpix.y:=z;
                      if abs((SubPix.y-(p1.y+m*(subpix.x-p1.x)))*genbuf)+0.71<h_curvwidth then begin //Abstand Pubpixel bis P1P2
                        hit:=true;
                        if Index=2147483647 then begin   //Wenn Pixel noch nicht erfasst wurde, wird der jetzt angelegt
                          Anteil:=AA_Total;
                          PixPlan[i-xMin_Arr,z]:=lPixArr;
                          index:=lPixArr;
                        end else begin                   //Wenn Pixel schon angelegt war, wird nur der Anteil auf max. gesetzt
                          Anteil:=AA_Total;
                        end;//wenn index <>-1
                      end;//if abs()...
                    end;//If h_Curvwidth>0.9
                    //-> Wenn der Pixel ausgehend von seiner Mitte nicht vollständig von Kurve bedeckt und nicht alle Subpixel erfasst
                    if (hit=false) and (Anteil<AA_Total) then begin
                      For k:=0 to (AA_hor_loc-1) do begin   //Äußere Subpixel-Schleife
                        Subpix.x:=i+AA_hor_startp+k*invAA_hor; //x-Subpixelposition berechnen (für ganze Spalte konstant)
                        ColumnStartIndex:=k*aa_ver_loc;     //Subpixelverschlüsselungs-Index für die Spalte
                        for n:=0 to (AA_ver_loc-1) do begin   //Spalte wird von oben nach unten durchgegangen
                          Subpix.y:=z+AA_ver_startp+n*invAA_ver;//y-Subpixelposition wird berechnet
                          if abs((SubPix.y-(p1.y+m*(subpix.x-p1.x)))*genbuf)<h_curvwidth then begin  // Wenn Subpixel innerhalb Kurve
                            if Index=2147483647 then begin //wenn dieser Pixel noch nicht angelegt ist
                              If AA_Total<33 then DSubPixBuf:=$1 else QSubPixBuf:=Qword($1);
                              Anteil:=1;
                              index:=lPixArr;
                            end else begin  //wenn ein Index zugewiesen war (Pixel war schon vorher angelegt)
                              If AA_Total<33 then begin
                                If GetBit(DSubPixBuf,ColumnStartIndex+n)=false then begin
                                  DSubPixBuf:=SetBit(DSubPixBuf,ColumnStartIndex+n);
                                  Inc(Anteil,1);
                                end;//wenn Subpixel noch nicht angelegt
                              end else begin
                                If QGetBit(QSubPixBuf,ColumnStartIndex+n)=false then begin
                                  QSubPixBuf:=QSetBit(QSubPixBuf,ColumnStartIndex+n);
                                  Inc(Anteil,1);
                                end;//wenn Subpixel noch nicht angelegt
                              end;
                            end;//if index >-1
                          end;//wenn Anteil <AA_Total
                        end;//AA_ver_loc
                      end;//AA_hor_loc
                    end;//if hit = false nach vereinfachung dicker Striche oder wenn dünner strich
                    If (Anteil>0)and(intbuf<Anteil) then begin //Wenn sich was verändert hat an diesem Pixel, und dieser durch min. 1 Subpixel erfasst wurde efolgt anlage und ggf. Array-vergrößerung
                      If PixArrMgr>PixArrMgrJump then begin    //intbuf wird verwendet um am Ende festzustellen, ob sich an einem Pixel der Anteil verändert hat
                        //-> Management für Array-Vergrößerungen alle PixArrMgrJump Pixelanlagen
                        Setlength(PixArray,lPixArr+PixArrMgrJump+1);
                        Setlength(Anteilplan,lPixArr+PixArrMgrJump+1);
                        if AA_Total<33 then Setlength(SubPixels,lPixArr+PixArrMgrJump+1)else Setlength(QSubPixels,lPixArr+PixArrMgrJump+1);
                        PixArrMgr:=0;
                      end;
                      If intbuf=0 then begin
                        PixArray[index].Punkt.x:=i;
                        PixArray[index].Punkt.y:=z;
                        PixPlan[i-xMin_Arr,z]:=index; //Wenn intbuf=0, dann erfolgte eine Neuanlage
                        inc(lPixArr,1);
                        inc(PixArrMgr,1);
                      end;
                      AnteilPlan[index]:=Anteil;
                      If AA_Total<33 then Subpixels[index]:=DSubPixBuf else QSubpixels[index]:=QSubPixBuf
                    end;
                  end;//Wenn aktuelles Pixel innerhalb der Auflösung liegt
                end;//endfor scanarea y-richtung
              end;//endfor ScanArea x-richtung
            end;//end Checks (wenn Scanrechteck max. 40 Pixel ausserhalb Bildschirm liegt und Errors=0)
            p1:=p2; //nach Durchlauf wird p1=p2 gesetzt und p2 neu ausgerechnet.
            if ComputationMode <>2 then Error1:=Error2;
          end;
          if ComputationMode<> 2 then xFunc:=xFunc+loc_StepWidth else Inc(CloudIndex,1);
        until (xFunc>xmax_loc)or(CloudIndex>=CloudSize); //endwhile xFunc<xmax

        if lPixArr>0 then begin
          setlength(PixArray,lPixArr);
          setlength(AnteilPlan,lPixArr);
          if AA_Total<33 then setlength(subpixels,lPixArr)else setlength(Qsubpixels,lPixArr);
        end;
        if length(PixArray)>0 then generate_Results;
      end;//if ana.error=0 begin

    Ende:=Time;
    stats.RenderTime:=round((ende-start)*24*60*60*1000)-stats.postprocess;
  end;

//=====================================================================================================================

  procedure tGraphCom.C_BufferedPointAACurv(const ComputationMode:byte);
   var
     Start,Ende:extended;
     p,subpix:tVec2D;
     ScanArea:trect;
     Error,index,lPixArr,yres_loc,xres_loc,Anteil,i,z,intbuf,xmin_arr,cloudIndex,CloudSize:longInt;
     loc_StepWidth,IntValue,xfunc,xMax_loc,yBuf,invAA_ver,invAA_hor,AA_ver_startp,AA_hor_startp,h_CurvWidth:double;
     hit:boolean;
     DSubPixBuf,DrawLimit:DWord;
     QSubPixBuf:Qword;
     ColumnStartIndex,k,n,AA_hor_loc,AA_ver_loc,AA_Total,PixArrMgr:byte;
   const
     PixArrMgrJump=128;    //PixArray wird alle PixArrMgrJump Schritte um PixArrMgrJump+1 vergrößert anstatt bei jedem Schritt um ein Feldelement, max. 254
    begin
     Start:=time;
      //---->> tAnalyse erzeugen und funktion zuweisen

        //---->> Berechnungsbedingungen setzen und abhängig vom Berechnungsmodus Graphberechnung starten
        If (ComputationMode=2)or((ana<>nil)and(Ana.Error=0)) then begin
          //--->> Zunächst für alle Computationmodes geltend
          PixArrMgr:=PixArrMgrJump+1; //Wenn PixArrMgr größer wird als PixArrMgrJump, dann wird das Array um PixArrMgrJump vergrößert (große Zeitvorteile)
          lPixArr:=length(PixArray);  //sollte folglich 0 sein, wenn erste funktion für die Klasse
          //--> Anti-Aliasing-Optionen setzen
          AA_hor_loc:=DrawSet.PerformanceSet.AA_Level_hor; AA_ver_loc:=DrawSet.PerformanceSet.AA_Level_ver; AA_Total:=AA_hor_loc*AA_ver_loc;
          h_CurvWidth:=DrawSet.PerformanceSet.CurvWidth/2; //halbe Kurvenbreite
          xmin_arr:=xMinThread_Canv;                       //Thread-bezogener Array-Offset für Canvas-x-Werte
          xres_loc:=ThreadResX;                            //Thread-bezogene Pixelauflösung in X
          yres_loc:=DrawSet.MainSizing.DrawResY;           //y-Res ist für alle Threads=Bildhöhe, da threads auf Ordinate sektioniert
          DrawLimit:=yres_loc+1500;                        //Wenn ein Wert 1500 Pixel unterhalb der Zeichenfläche ist. Alles noch tiefer wird ignoriert und auf DrawLimit gesetzt
          CloudSize:=length(PCloud);                       //Größe für Punktewolke (falls vorhanden) ermitteln
          CloudIndex:=-1;                                  //Fortschritts-Index innerhalb der Punktewolke zurücksetzen
          If (ThreadNumber<>DrawSet.PerformanceSet.ThreadCount-1)  or  ((ThreadNumber=DrawSet.PerformanceSet.ThreadCount-1)and(IntMax=DrawSet.Kartesianset.xmax)) then xMax_loc:=xMaxThread+(h_Curvwidth/zoomx*3)
            else xMax_loc:=xMaxThread;
          //--> Oft gebrauchte, über die gesamte Thread-Dauer konstante bleibende Werte ausrechnen, ums nicht 10^7 mal berechnen zu müssen
          invAA_ver:=1/AA_ver_loc;
          invAA_hor:=1/AA_hor_loc;
          AA_ver_startp:=-0.5+0.5/AA_ver_loc;  //Vertikaler Subpixeloffset oben
          AA_hor_startp:=-0.5+0.5/AA_hor_loc;  //horizontaler Subpixeloffset links
          //--> Wenn kein Punktewolkenmodus
          If ComputationMode<>2 then begin
            //-> Startpunkte berechnen
            loc_StepWidth:=DrawSet.PerformanceSet.unsharpnes/zoomx;
            If ComputationMode=0 then begin
              //-> Zeichenvorgang einer "normalen" Funktion
              If (ThreadNumber>0)  or  ((ThreadNumber=0)and(IntMin=DrawSet.Kartesianset.xmin)) then  xfunc:=xMinThread-(h_Curvwidth/zoomx*3) //Um sichtbare Übergänge zwischen den Threads zu entfernen, muss man min. 1,5 Strichbeiten am Anfang und Ende zu viel Berechnen pro Thread
                else xfunc:=xMinThread;
            end else begin
              //-> Zeichenvorgang einer numerischen Integral-Funktion
              xFunc:=xMinThread; //im Modus für numerische Integrale darf man nicht zurücksetzen.
              IntValue:=IndividualOffset;
            end;
          end else begin //-> Wenn Punktewolken-Modus
            CloudIndex:=0;
            ybuf:=0;
            error:=0; //Wenn Error <>0 dann, werden keine Ausgaben für einen Punkt berechnet. Error braucht man nur für richtige Funktionen
            xFunc:=xMinThread;
          end;
          //---->> Hauptwiederholschleife
          Repeat
            //---> Neuen P2 ausrechnen
            Case ComputationMode of
              0: begin                                 //normaler Berechnungsmodus
                p.x:=xRealToCanv(xfunc);
                ybuf:=yRealToCanv(Ana.FWert(xFunc));
                Error:=ana.Error;
              end;
              1: begin                                 //numerischer Integrationsmodus
                p.x:=xRealToCanv(xfunc);
                IntValue:=IntValue+(Ana.FWert(xFunc+loc_stepwidth/2)*loc_Stepwidth);
                ybuf:=yRealToCanv(intValue);
                Error:=ana.Error;
              end;
              2: begin                                 //Punktewolkenmodus
                p.x:=xRealToCanv(PCloud[CloudIndex].x);
                ybuf:=yRealToCanv(PCloud[CloudIndex].y);
              end;
            end;
            if ybuf>DrawLimit then ybuf:=DrawLimit;
            if ybuf<-1500 then ybuf:=-1500;
            p.y:=ybuf;
            //If CloudIndex>1 then showmessage('Index'+' '+IntToStr(CloudIndex)+'x '+ FloatToStr(PCloud[CloudIndex].x)+' '+FloatToStr(p.x)+'  y'+FloatToStr(PCloud[CloudIndex].x)+' '+FloatToStr(p.y));
              //---> Wenn P bis max. 40 Pix ausserhalb des sichtbaren Bereichs, werden noch Scanbereich erfasst, und evtl. Pixel, falls sichtbar, berechnet
              If (p.y<yres_loc+40)and(p.y>-41)and(error=0) then begin
              //-> Scan-Rechteck festlegen
              ScanArea.Left:=round(p.x-h_curvwidth);    //Betroffene Pixel in x-Richtung abstecken
              ScanArea.Right:=round(p.x+h_curvwidth);   //Betroffene Pixel in x-Richtung abstecken
              ScanArea.Bottom:=round(p.y-h_curvwidth);  //Betroffene Pixel in y-Richtung abstecken
              ScanArea.Top:=round(p.y+h_curvwidth);     //Betroffene Pixel in y-Richtung abstecken

              for i:=ScanArea.Left to ScanArea.Right do begin //i entspricht x-Richtung für Pixel, Pixelebene
                for z:=ScanArea.bottom to ScanArea.Top do begin  //z entspricht y-Richtung für Pixel, Pixelebene
                  If (i-xMin_Arr>-1)and (i-xMin_Arr<xRes_loc) and(z>-1)and(z<yres_loc) then begin  //Wenn aktuelles Pixel im sichtbaren Bereich, wird auf niedrigerer Ebene weitergerechnet
                  hit:=false;
                  //-> Index für aktuelles Pixel im Ergebnis-Array wird aus der PixMap ausgelesen
                  Index:=PixPlan[i-xMin_Arr,z];
                  If Index <2147483647 then begin // Index leer=2147483647, <2147483647 heißt, der Pixel wurde schonmal angelegt an der durch die Nummer angezeigten Adresse
                    Anteil:=AnteilPlan[index];
                    intbuf:=anteil;    //intbuf wird verwendet um am Ende festzustellen, ob sich an einem Pixel der Anteil verändert hat
                    if Intbuf<AA_Total then If AA_Total<33 then DSubPixBuf:=Subpixels[index] else QSubPixBuf:=QSubpixels[index];
                  end else begin                  // Pixel wird mit Anteil=0 neu angelegt, falls Index=high(integer)
                    Anteil:=0;
                    intbuf:=0;
                    If AA_Total<33 then DSubPixBuf:=$0 else QSubPixBuf:=Qword($0);
                  end;
                  //-> kommt bei dicken Strichen zum Einsatz, wenn Pixel nicht bereits vollen Farbanteil hat
                  if (h_curvwidth>0.9)and(anteil<AA_Total) then begin
                    Subpix.x:=i;
                    Subpix.y:=z;
                    if sqrt(sqr(SubPix.y-p.y)+sqr(SubPix.x-p.x))+0.71<h_curvwidth then begin //Abstand Pubpixel bis P1P2
                      hit:=true;
                      if Index=2147483647 then begin   //Wenn Pixel noch nicht erfasst wurde, wird der jetzt angelegt
                        Anteil:=AA_Total;
                        PixPlan[i-xMin_Arr,z]:=lPixArr;
                        index:=lPixArr;
                      end else begin                   //Wenn Pixel schon angelegt war, wird nur der Anteil auf max. gesetzt
                        Anteil:=AA_Total;
                      end;//wenn index <>-1
                    end;//if abs()...
                  end;//If h_Curvwidth>0.9
                  //-> Wenn der Pixel ausgehend von seiner Mitte nicht vollständig von Kurve bedeckt und nicht alle Subpixel erfasst
                  if (hit=false) and (Anteil<AA_Total) then begin
                    For k:=0 to (AA_hor_loc-1) do begin   //Äußere Subpixel-Schleife
                      Subpix.x:=i+AA_hor_startp+k*invAA_hor; //x-Subpixelposition berechnen (für ganze Spalte konstant)
                      ColumnStartIndex:=k*aa_ver_loc;     //Subpixelverschlüsselungs-Index für die Spalte
                      for n:=0 to (AA_ver_loc-1) do begin   //Spalte wird von oben nach unten durchgegangen
                        Subpix.y:=z+AA_ver_startp+n*invAA_ver;//y-Subpixelposition wird berechnet
                        if sqrt(sqr(SubPix.y-p.y)+sqr(SubPix.x-p.x))<h_curvwidth then begin  // Wenn Subpixel innerhalb Kurve
                          if Index=2147483647 then begin //wenn dieser Pixel noch nicht angelegt ist
                            If AA_Total<33 then DSubPixBuf:=$1 else QSubPixBuf:=Qword($1);
                            Anteil:=1;
                            index:=lPixArr;
                          end else begin  //wenn ein Index zugewiesen war (Pixel war schon vorher angelegt)
                            If AA_Total<33 then begin
                              If GetBit(DSubPixBuf,ColumnStartIndex+n)=false then begin
                                DSubPixBuf:=SetBit(DSubPixBuf,ColumnStartIndex+n);
                                Inc(Anteil,1);
                              end;//wenn Subpixel noch nicht angelegt
                            end else begin
                              If QGetBit(QSubPixBuf,ColumnStartIndex+n)=false then begin
                                QSubPixBuf:=QSetBit(QSubPixBuf,ColumnStartIndex+n);
                                Inc(Anteil,1);
                              end;//wenn Subpixel noch nicht angelegt
                            end;
                          end;//if index >-1
                        end;//wenn Anteil <AA_Total
                      end;//AA_ver_loc
                    end;//AA_hor_loc
                  end;//if hit = false nach vereinfachung dicker Striche oder wenn dünner strich
                  If (Anteil>0)and(intbuf<Anteil) then begin //Wenn sich was verändert hat an diesem Pixel, und dieser durch min. 1 Subpixel erfasst wurde efolgt anlage und ggf. Array-vergrößerung
                    If PixArrMgr>PixArrMgrJump then begin    //intbuf wird verwendet um am Ende festzustellen, ob sich an einem Pixel der Anteil verändert hat
                      //-> Management für Array-Vergrößerungen alle PixArrMgrJump Pixelanlagen
                      Setlength(PixArray,lPixArr+PixArrMgrJump+1);
                      Setlength(Anteilplan,lPixArr+PixArrMgrJump+1);
                      if AA_Total<33 then Setlength(SubPixels,lPixArr+PixArrMgrJump+1)else Setlength(QSubPixels,lPixArr+PixArrMgrJump+1);
                      PixArrMgr:=0;
                    end;
                    If intbuf=0 then begin
                      PixArray[index].Punkt.x:=i;
                      PixArray[index].Punkt.y:=z;
                      PixPlan[i-xMin_Arr,z]:=index; //Wenn intbuf=0, dann erfolgte eine Neuanlage
                      inc(lPixArr,1);
                      inc(PixArrMgr,1);
                    end;
                    AnteilPlan[index]:=Anteil;
                    If AA_Total<33 then Subpixels[index]:=DSubPixBuf else QSubpixels[index]:=QSubPixBuf
                  end;
                end;//Wenn aktuelles Pixel innerhalb der Auflösung liegt
              end;//endfor scanarea y-richtung
            end;//endfor ScanArea x-richtung
          end;//end Checks (wenn Scanrechteck max. 40 Pixel ausserhalb Bildschirm liegt und Errors=0)
          if ComputationMode<> 2 then xFunc:=xFunc+loc_StepWidth else Inc(CloudIndex,1);
        until (xFunc>xmax_loc)or(CloudIndex>=CloudSize); //endwhile xFunc<xmax

        if lPixArr>0 then begin
          setlength(PixArray,lPixArr);
          setlength(AnteilPlan,lPixArr);
          if AA_Total<33 then setlength(subpixels,lPixArr)else setlength(Qsubpixels,lPixArr);
        end;
        if length(PixArray)>0 then generate_Results;
      end;//if ana.error=0 begin



    Ende:=Time;
    stats.RenderTime:=round((ende-start)*24*60*60*1000)-stats.postprocess;
  end;

//=====================================================================================================================

  procedure tGraphCom.C_BufferedRoundRectAACurv(const ComputationMode:byte);
   var
     Start,Ende:extended;
     p1,p2,p1h,p1t,p2h,p2t,p1p2,pm,subpix:tVec2D;
     ScanArea:trect;
     Error1,Error2,index,lPixArr,yres_loc,xres_loc,Anteil,i,z,intbuf,xmin_arr,cloudIndex,CloudSize:longInt;
     loc_StepWidth,xfunc,xMax_loc,yBuf,IntValue,invAA_ver,invAA_hor,AA_ver_startp,AA_hor_startp,h_CurvWidth,genbuf,m,beta:double;
     hit,CircSwitch,SwitchMoment:boolean;
     DSubPixBuf,DrawLimit:DWord;
     QSubPixBuf:Qword;
     ColumnStartIndex,k,n,AA_hor_loc,AA_ver_loc,AA_Total,PixArrMgr:byte;
   const
     PixArrMgrJump=128;    //PixArray wird alle PixArrMgrJump Schritte um PixArrMgrJump+1 vergrößert anstatt bei jedem Schritt um ein Feldelement, max. 254
    begin
     Start:=time;
      //---->> tAnalyse erzeugen und funktion zuweisen

        //---->> Berechnungsbedingungen setzen und abhängig vom Berechnungsmodus Graphberechnung starten
        If (ComputationMode=2)or((ana<>nil)and(Ana.Error=0)) then begin
          //--->> Zunächst für alle Computationmodes geltend
          PixArrMgr:=PixArrMgrJump+1; //Wenn PixArrMgr größer wird als PixArrMgrJump, dann wird das Array um PixArrMgrJump vergrößert (große Zeitvorteile)
          lPixArr:=length(PixArray);  //sollte folglich 0 sein, wenn erste funktion für die Klasse
          //--> Anti-Aliasing-Optionen setzen
          AA_hor_loc:=DrawSet.PerformanceSet.AA_Level_hor; AA_ver_loc:=DrawSet.PerformanceSet.AA_Level_ver; AA_Total:=AA_hor_loc*AA_ver_loc;
          h_CurvWidth:=DrawSet.PerformanceSet.CurvWidth/2; //halbe Kurvenbreite
          xmin_arr:=xMinThread_Canv;                       //Thread-bezogener Array-Offset für Canvas-x-Werte
          xres_loc:=ThreadResX;                            //Thread-bezogene Pixelauflösung in X
          yres_loc:=DrawSet.MainSizing.DrawResY;           //y-Res ist für alle Threads=Bildhöhe, da threads auf Ordinate sektioniert
          DrawLimit:=yres_loc+1500;                        //Wenn ein Wert 1500 Pixel unterhalb der Zeichenfläche ist. Alles noch tiefer wird ignoriert und auf DrawLimit gesetzt
          CloudSize:=length(PCloud);                       //Größe für Punktewolke (falls vorhanden) ermitteln
          CloudIndex:=-1;                                  //Fortschritts-Index innerhalb der Punktewolke zurücksetzen
          If (ThreadNumber<>DrawSet.PerformanceSet.ThreadCount-1)  or  ((ThreadNumber=DrawSet.PerformanceSet.ThreadCount-1)and(IntMax=DrawSet.Kartesianset.xmax)) then xMax_loc:=xMaxThread+(h_Curvwidth/zoomx*3)
            else xMax_loc:=xMaxThread;
          //--> Oft gebrauchte, über die gesamte Thread-Dauer konstante bleibende Werte ausrechnen, ums nicht 10^7 mal berechnen zu müssen
          invAA_ver:=1/AA_ver_loc;
          invAA_hor:=1/AA_hor_loc;
          AA_ver_startp:=-0.5+0.5/AA_ver_loc;  //Vertikaler Subpixeloffset oben
          AA_hor_startp:=-0.5+0.5/AA_hor_loc;  //horizontaler Subpixeloffset links
          //--> Wenn kein Punktewolkenmodus
          If ComputationMode<>2 then begin
            //-> Startpunkte berechnen
            loc_StepWidth:=DrawSet.PerformanceSet.unsharpnes/zoomx;
            If ComputationMode=0 then begin
              //-> Zeichenvorgang einer "normalen" Funktion
              If (ThreadNumber>0)  or  ((ThreadNumber=0)and(IntMin=DrawSet.Kartesianset.xmin)) then  xfunc:=xMinThread-(h_Curvwidth/zoomx*3) //Um sichtbare Übergänge zwischen den Threads zu entfernen, muss man min. 1,5 Strichbeiten am Anfang und Ende zu viel Berechnen pro Thread
                else xfunc:=xMinThread; //Um sichtbare Übergänge zwischen den Threads zu entfernen, muss man min. 1,5 Strichbeiten am Anfang und Ende zu viel Berechnen pro Thread
              p1.x:=xRealToCanv(xfunc);
              ybuf:=yRealToCanv(Ana.FWert(xFunc));
            end else begin
              //-> Zeichenvorgang einer "normalen" Funktion
              xFunc:=xMinThread; //im Modus für numerische Integrale darf man nicht zurücksetzen.
              p1.x:=xRealToCanv(xfunc);
              IntValue:=IndividualOffset;
              ybuf:=yRealToCanv(IndividualOffset);
            end;
            //---> yBuf in Canvas-Koordinaten macht keinen sichtbaren Unterschied, ob -10E15 oder  -1500
            if ybuf>DrawLimit then ybuf:=DrawLimit;
            if ybuf<-1500 then ybuf:=-1500;
            p1.y:=ybuf;
            Error1:=Ana.Error;
            xFunc:=xFunc+loc_StepWidth;
          end else begin //-> Wenn Punktewolken-Modus
            CloudIndex:=1;
            error1:=0;error2:=0;//Wenn Error <>0 dann, werden keine Ausgaben für einen Punkt berechnet. Error braucht man nur für richtige Funktionen
            p1.x:=xRealToCanv(PCloud[0].x);
            P1.y:=yRealToCanv(PCloud[0].y);
            ybuf:=0;
            xFunc:=xMinThread;
          end;
          SwitchMoment:=false;
          CircSwitch:=false;
          //---->> Hauptwiederholschleife
          Repeat
            //---> Neuen P2 ausrechnen
            If CircSwitch=false then begin
              If ((ComputationMode<>2)and(xFunc>=xMinThread+(h_Curvwidth/zoomx)))  or  ((ComputationMode=2)and(CloudIndex>=1)) then begin
                CircSwitch:=true;
                SwitchMoment:=true;
              end;
            end;
            Case ComputationMode of
              0: begin                                 //normaler Berechnungsmodus
                p2.x:=xRealToCanv(xfunc);
                ybuf:=yRealToCanv(Ana.FWert(xFunc));
                Error2:=ana.Error;
              end;
              1: begin                                 //numerischer Integrationsmodus
                p2.x:=xRealToCanv(xfunc);
                IntValue:=IntValue+(Ana.FWert(xFunc+loc_stepwidth/2)*loc_Stepwidth);
                ybuf:=yRealToCanv(intValue);
                Error2:=ana.Error;
              end;
              2: begin                                 //Punktewolkenmodus
                p2.x:=xRealToCanv(PCloud[CloudIndex].x);
                ybuf:=yRealToCanv(PCloud[CloudIndex].y);
              end;
            end;
            if ybuf>DrawLimit then ybuf:=DrawLimit;
            if ybuf<-1500 then ybuf:=-1500;
            p2.y:=ybuf;
            //---> Wenn am Bildschirm überhaupt eine sichtbare Veränderung eintritt, dann muss der ganze AA-Algorithmus durchlaufen werden und anschließend wird der P1 der alte P2. Der neue P2 wird berechnet
            If (abs(p1.y-p2.y)>0.67)or(p2.x-p1.x>0.67)or(h_CurvWidth>1.95) then begin  //Bei dicken Kurven muss der Alg. immer durchlaufen werden, da sonst komische Ausbrüche bei Extremstellen hochfreq. Funktionen
              p1p2.x:=p2.x-p1.x; p1p2.y:=p2.y-p1.y;  //Richtungsvektor P1 nach P2
              m:=(p1p2.y)/(p1p2.x);
              beta:=arctan(m)+pi/2;
              //-> Rechteckecken bei P2 in Y
              genbuf:=sin(beta)*h_Curvwidth;
              p2h.y:=p2.y+genbuf;
              p2t.y:=p2.y-genbuf;
              //-> Rechteckecken bei P2 in X
              genbuf:=cos(beta)*h_Curvwidth;
              p2h.x:=p2.x+genbuf;       //prefetch von h_CurvWidth
              p2t.x:=p2.x-genbuf;
              //-> Rechteckecken bei P1 in X und Y
              p1h.x:=p2h.x-(p1p2.x); p1h.y:=p2h.y-(p1p2.y);
              p1t.x:=p2t.x-(p1p2.x); p1t.y:=p2t.y-(p1p2.y);
              pm.x:=(p2h.x+p1h.x)/2;
              pm.y:=(p2h.y+p1h.y)/2;
              //---> Wenn P1 oder P2 bis max. 40 Pix ausserhalb des sichtbaren Bereichs, werden noch Scanbereich erfasst, und evtl. Pixel, falls sichtbar, berechnet
              If (((p1.y<yres_loc+40)and(p1.y>-41))or((p2.y<yres_loc+40)and(p2.y>-41)))and((error1=0) and (Error2=0)) then begin
              //-> Scan-Rechteck festlegen
                If p1.y<p2.y then begin
                  if (not CircSwitch)or(Switchmoment) then ScanArea.Left:=round(p1.x-h_curvwidth) else ScanArea.Left:=round(p1h.x);   //Betroffene Pixel in x-Richtung abstecken
                  if (not CircSwitch)and(not Switchmoment) then ScanArea.Right:=round(p2t.x) else ScanArea.Right:=round(p2.x+h_CurvWidth);   //Betroffene Pixel in x-Richtung abstecken
                  if (not CircSwitch)or(Switchmoment) then ScanArea.Bottom:=round(p1.y-h_curvwidth) else ScanArea.Bottom:=round(p1t.y);  //Betroffene Pixel in y-Richtung abstecken
                  if (not CircSwitch)and(not Switchmoment) then ScanArea.Top:=round(p2h.y) else ScanArea.Top:=round(p2.y+h_CurvWidth);     //Betroffene Pixel in y-Richtung abstecken
                end else begin
                  if (not CircSwitch)or(Switchmoment) then ScanArea.left:=round(p1.x-h_curvWidth) else ScanArea.Left:=round(p1t.x);    //Betroffene Pixel in x-Richtung abstecken
                  if (not CircSwitch)and(not Switchmoment) then ScanArea.Right:=round(p2h.x) else ScanArea.Right:=round(p2.x+h_CurvWidth);   //Betroffene Pixel in x-Richtung abstecken
                  if (not CircSwitch)and(not Switchmoment) then ScanArea.Bottom:=round(p2t.y) else ScanArea.Bottom:=round(p2.y-h_CurvWidth);  //Betroffene Pixel in y-Richtung abstecken
                  if (not CircSwitch)or(Switchmoment) then ScanArea.Top:=round(p1.y+h_CurvWidth) else ScanArea.Top:=round(p1h.y);     //Betroffene Pixel in y-Richtung abstecken
                end;
                genbuf:=1/sqrt(sqr(m)+1); //=cos(arctan(m)); Braucht man für Abstandsberechnungen der Subpixel zur Linie P1P2 (umrechnung delta_Y in orthogonal-Höhe)

                for i:=ScanArea.Left to ScanArea.Right do begin //i entspricht x-Richtung für Pixel, Pixelebene
                  for z:=ScanArea.bottom to ScanArea.Top do begin  //z entspricht y-Richtung für Pixel, Pixelebene
                    If (i-xMin_Arr>-1)and (i-xMin_Arr<xRes_loc) and(z>-1)and(z<yres_loc) then begin  //Wenn aktuelles Pixel im sichtbaren Bereich, wird auf niedrigerer Ebene weitergerechnet
                    hit:=false;
                    //-> Index für aktuelles Pixel im Ergebnis-Array wird aus der PixMap ausgelesen
                    Index:=PixPlan[i-xMin_Arr,z];
                    If Index <2147483647 then begin // Index leer=2147483647, <2147483647 heißt, der Pixel wurde schonmal angelegt an der durch die Nummer angezeigten Adresse
                      Anteil:=AnteilPlan[index];
                      intbuf:=anteil;    //intbuf wird verwendet um am Ende festzustellen, ob sich an einem Pixel der Anteil verändert hat
                      if Intbuf<AA_Total then If AA_Total<33 then DSubPixBuf:=Subpixels[index] else QSubPixBuf:=QSubpixels[index];
                    end else begin                  // Pixel wird mit Anteil=0 neu angelegt, falls Index=high(integer)
                      Anteil:=0;
                      intbuf:=0;
                      If AA_Total<33 then DSubPixBuf:=$0 else QSubPixBuf:=Qword($0);
                    end;
                    //-> kommt bei dicken Strichen zum Einsatz, wenn Pixel nicht bereits vollen Farbanteil hat
                    if (h_curvwidth>0.9)and(anteil<AA_Total) then begin
                      Subpix.x:=i;
                      Subpix.y:=z;
                      if  ((CircSwitch)and(sqrt(sqr(subpix.x-p2.x)+sqr(subpix.y-p2.y))+0.71<h_curvwidth))  or  (((not CircSwitch)or(SwitchMoment))and(sqrt(sqr(subpix.x-p1.x)+sqr(subpix.y-p1.y))+0.71<h_curvwidth))  or  ((abs((SubPix.y-(p1.y+m*(subpix.x-p1.x)))*genbuf)+0.71<h_curvwidth)and(abs((-SubPix.x-(-pm.x+m*(subpix.y-pm.y)))*genbuf)+0.71<(p2.x-p1.x)/(2*genbuf))) then begin //Abstand Pubpixel bis P1P2
                        hit:=true;
                        if Index=2147483647 then begin   //Wenn Pixel noch nicht erfasst wurde, wird der jetzt angelegt
                          Anteil:=AA_Total;
                          PixPlan[i-xMin_Arr,z]:=lPixArr;
                          index:=lPixArr;
                        end else begin                   //Wenn Pixel schon angelegt war, wird nur der Anteil auf max. gesetzt
                          Anteil:=AA_Total;
                        end;//wenn index <>-1
                      end;//if abs()...
                    end;//If h_Curvwidth>0.9
                    //-> Wenn der Pixel ausgehend von seiner Mitte nicht vollständig von Kurve bedeckt und nicht alle Subpixel erfasst
                    if (hit=false) and (Anteil<AA_Total) then begin
                      For k:=0 to (AA_hor_loc-1) do begin   //Äußere Subpixel-Schleife
                        Subpix.x:=i+AA_hor_startp+k*invAA_hor; //x-Subpixelposition berechnen (für ganze Spalte konstant)
                        ColumnStartIndex:=k*aa_ver_loc;     //Subpixelverschlüsselungs-Index für die Spalte
                        for n:=0 to (AA_ver_loc-1) do begin   //Spalte wird von oben nach unten durchgegangen
                          Subpix.y:=z+AA_ver_startp+n*invAA_ver;//y-Subpixelposition wird berechnet
                          if  ((CircSwitch)and(sqrt(sqr(subpix.x-p2.x)+sqr(subpix.y-p2.y))<h_curvwidth))  or  (((not CircSwitch)or(SwitchMoment))and(sqrt(sqr(subpix.x-p1.x)+sqr(subpix.y-p1.y))<h_curvwidth))  or  ((abs((SubPix.y-(p1.y+m*(subpix.x-p1.x)))*genbuf)<h_curvwidth)and(abs((-SubPix.x-(-pm.x+m*(subpix.y-pm.y)))*genbuf)<(p2.x-p1.x)/(2*genbuf))) then begin  // Wenn Subpixel innerhalb Kurve
                            if Index=2147483647 then begin //wenn dieser Pixel noch nicht angelegt ist
                              If AA_Total<33 then DSubPixBuf:=$1 else QSubPixBuf:=Qword($1);
                              Anteil:=1;
                              index:=lPixArr;
                            end else begin  //wenn ein Index zugewiesen war (Pixel war schon vorher angelegt)
                              If AA_Total<33 then begin
                                If GetBit(DSubPixBuf,ColumnStartIndex+n)=false then begin
                                  DSubPixBuf:=SetBit(DSubPixBuf,ColumnStartIndex+n);
                                  Inc(Anteil,1);
                                end;//wenn Subpixel noch nicht angelegt
                              end else begin
                                If QGetBit(QSubPixBuf,ColumnStartIndex+n)=false then begin
                                  QSubPixBuf:=QSetBit(QSubPixBuf,ColumnStartIndex+n);
                                  Inc(Anteil,1);
                                end;//wenn Subpixel noch nicht angelegt
                              end;
                            end;//if index >-1
                          end;//wenn Anteil <AA_Total
                        end;//AA_ver_loc
                      end;//AA_hor_loc
                    end;//if hit = false nach vereinfachung dicker Striche oder wenn dünner strich
                    If (Anteil>0)and(intbuf<Anteil) then begin //Wenn sich was verändert hat an diesem Pixel, und dieser durch min. 1 Subpixel erfasst wurde efolgt anlage und ggf. Array-vergrößerung
                      If PixArrMgr>PixArrMgrJump then begin    //intbuf wird verwendet um am Ende festzustellen, ob sich an einem Pixel der Anteil verändert hat
                        //-> Management für Array-Vergrößerungen alle PixArrMgrJump Pixelanlagen
                        Setlength(PixArray,lPixArr+PixArrMgrJump+1);
                        Setlength(Anteilplan,lPixArr+PixArrMgrJump+1);
                        if AA_Total<33 then Setlength(SubPixels,lPixArr+PixArrMgrJump+1)else Setlength(QSubPixels,lPixArr+PixArrMgrJump+1);
                        PixArrMgr:=0;
                      end;
                      If intbuf=0 then begin
                        PixArray[index].Punkt.x:=i;
                        PixArray[index].Punkt.y:=z;
                        PixPlan[i-xMin_Arr,z]:=index; //Wenn intbuf=0, dann erfolgte eine Neuanlage
                        inc(lPixArr,1);
                        inc(PixArrMgr,1);
                      end;
                      AnteilPlan[index]:=Anteil;
                      If AA_Total<33 then Subpixels[index]:=DSubPixBuf else QSubpixels[index]:=QSubPixBuf
                    end;
                  end;//Wenn aktuelles Pixel innerhalb der Auflösung liegt
                end;//endfor scanarea y-richtung
              end;//endfor ScanArea x-richtung
            end;//end Checks (wenn Scanrechteck max. 40 Pixel ausserhalb Bildschirm liegt und Errors=0)
            p1:=p2; //nach Durchlauf wird p1=p2 gesetzt und p2 neu ausgerechnet.
            if ComputationMode <>2 then Error1:=Error2;
          end;
          If SwitchMoment=true then SwitchMoment:=false;
          if ComputationMode<> 2 then xFunc:=xFunc+loc_StepWidth else Inc(CloudIndex,1);
        until (xFunc>xmax_loc)or(CloudIndex>=CloudSize); //endwhile xFunc<xmax

        if lPixArr>0 then begin
          setlength(PixArray,lPixArr);
          setlength(AnteilPlan,lPixArr);
          if AA_Total<33 then setlength(subpixels,lPixArr)else setlength(Qsubpixels,lPixArr);
        end;
        if length(PixArray)>0 then generate_Results;
      end;//if ana.error=0 begin



    Ende:=Time;
    stats.RenderTime:=round((ende-start)*24*60*60*1000)-stats.postprocess;
  end;

//=====================================================================================================================

  procedure tGraphCom.Execute;
    begin
      If ThreadIsValid then begin
        try
          If GraphCompMode<>2 then begin //tAnalyse initialisieren, wenn keine Punktewolken gezeichnet werden sollen
            Ana:=tAnalyse.create;
            Ana.FEing(funktion);
          end;
          If (DrawSet.PerformanceSet.unsharpnes>0)or(DrawSet.PerformanceSet.AA_Mode=1) then begin
            case DrawSet.PerformanceSet.AA_Mode of   //Start der entsprechenden AA-Verfahren je nach Wahl
              1: C_unbufferedTetragonAACurv(GraphCompMode);
              2: C_BufferedHexagonAACurv(GraphCompMode);
              3: C_BufferedPointAACurv(GraphCompMode);
              4: C_BufferedRoundRectAACurv(GraphCompMode);
            end;
          end;
        finally
          If Ana<>nil then begin       //tAnalyse abschließen, falls diese initialisiert wurde.
            Ana.Free;
            ana:=nil;
          end;
        end;
      end;
    end;

//=====================================================================================================================

end.

