unit UPointEdit;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, LResources, Forms, Controls, Graphics, Dialogs,
  Grids, Buttons, StdCtrls, Menus, uGraphcom, u_prm_numerik;

type

  { TFPointEdit }

  TFPointEdit = class(TForm)
    CB_ordered: TCheckBox;
    CB_ctype: TComboBox;
    Label1: TLabel;
    Label2: TLabel;
    PM_PEdit_sortView: TMenuItem;
    PM_PEdit_delRow: TMenuItem;
    PM_PEdit_newRow: TMenuItem;
    PM_PEdit: TPopupMenu;
    SB_reset: TSpeedButton;
    SB_Ready: TSpeedButton;
    SG_PEdit: TStringGrid;
    SpeedButton1: TSpeedButton;
    procedure CB_orderedChange(Sender: TObject);
    procedure FormClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure PM_PEditClose(Sender: TObject);
    procedure PM_PEditPopup(Sender: TObject);
    procedure PM_PEdit_delRowClick(Sender: TObject);
    procedure PM_PEdit_newRowClick(Sender: TObject);
    procedure PM_PEdit_sortViewClick(Sender: TObject);
    procedure SB_ReadyClick(Sender: TObject);
    procedure SG_PEditMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure SG_PEditSelectCell(Sender: TObject; aCol, aRow: Integer;
      var CanSelect: Boolean);
    procedure SpeedButton1Click(Sender: TObject);
  private
    ServiceCaller:byte;                             //0=FMain->Cb_Mainfunc; 1=FFuncEdit->LE_FuncText; 2=FEinstellungen->CB_PrevCurv
    CloudStr:String;
    Cloud:tPointCloud;
    Numerik:tNumerik;
    enableOnSelect:boolean;                      //Wenn man nicht möchte, dass einem OnSelectCell immer in die Quäre kommt, kann man die Var zeitw. auf false setzen. Aber das Zurücksetzen nicht vergessen!
    procedure InitGui(loc_CloudStr:string;var ref_cloud:tPointCloud);
    function GuiToPointString(forceSort:boolean):String;
    procedure refreshCounter;                    //neu durchnummerieren
    procedure deleteEmptyRows(silent:boolean);   //kann alle leeren Rows entfernen
    { private declarations }
  public
    constructor create (ServiceCallerImp:byte; CloudStrImp:String);
    Destructor free;
    { public declarations }
  end; 

var
  FPointEdit: TFPointEdit;

implementation

uses uMain, uFuncEdit, uEinstellungen;

{ TFPointEdit }

Constructor TFPointEdit.create(ServiceCallerImp:byte; CloudStrImp:String);  //0=FMain->Cb_Mainfunc; 1=FFuncEdit->LE_FuncText; 2=FEinstellungen->CB_PrevCurv
  begin
    inherited Create(nil);
    ServiceCaller:=ServiceCallerImp;
    CloudStr:=CloudStrImp;
    enableOnSelect:=true;
    setlength(Cloud,0);
    Numerik:=tNumerik.create;
  end;

Destructor TFPointEdit.free;
  begin
    Cloud:=nil;
    Numerik.free;
    inherited destroy;
  end;

procedure tFPointEdit.refreshCounter;
 var
   i:longint;
  begin
    for i:=1 to SG_PEdit.RowCount-1 do begin   //Nummerierung wird neu generiert.
      SG_PEdit.Cells[0,i]:=IntToStr(i);
    end;
  end;

procedure tFPointEdit.deleteEmptyRows(silent:boolean);
 var
   i,brokenrow:longInt;
   buf:string='';
  begin
    enableOnSelect:=false;
    brokenrow:=0;
    //SG_PEdit.Options:=SG_PEdit.Options-[goRowSelect];
    SG_PEdit.Col:=1; SG_Pedit.row:=1;  //Die Selection für das Grid auf 1,1 setzen (zu diesem Zeitpunkt muss OnSelectCell durch enableOnSelect:=false deaktiviert sein)
    i:=SG_PEdit.RowCount-1;
    while i>0 do begin                 //Leere Zeilen werden entfernt
      If (SG_PEdit.Cells[1,i]='') and (SG_PEdit.Cells[2,i]='') then SG_PEdit.DeleteColRow(false,i)
        else If (SG_PEdit.Cells[1,i]='') xor (SG_PEdit.Cells[2,i]='') then inc(BrokenRow,1);
      dec(i,1);
    end;
    SG_PEdit.RowCount:=SG_PEdit.RowCount+1;
    refreshCounter;   //Nummerierung wird neu generiert.
    if brokenrow>0 then begin
      if brokenrow>1 then buf:='n';
      if not silent then MessageDlg('*Info*','Es gibt '+ IntToStr(Brokenrow) +' unvollständige Zeile'+Buf+'. Bleibt diese unvollständig, wird sie nicht weiter berücksichtigt.', mtinformation, [mbOk] , 0);
    end;
    enableOnSelect:=true;
  end;

Procedure TFPointEdit.InitGui(loc_CloudStr:string;var ref_cloud:tPointCloud);
 var
   i:int64=0;
   ordered,ctype:char;
  begin
    SG_PEdit.Col:=1; SG_Pedit.row:=1;  //Die Selection für das Grid auf 1,1 setzen
    SG_PEdit.Clean([gzNormal]);
    If (Pos('pointcloud(',LowerCase(loc_cloudstr))>0) and (length(loc_CloudStr)>14) then begin
      Numerik.StrToPCloud(loc_CloudStr,ref_Cloud,ordered,ctype);
      if ordered='s' then CB_ordered.Checked:=true else CB_ordered.Checked:=false;
      case ctype of
        'n':CB_ctype.ItemIndex:=0;
        'l':CB_ctype.ItemIndex:=1;
      end;
    end;
    enableOnSelect:=false;
    SG_PEdit.RowCount:=length(ref_Cloud)+2;
    SG_PEdit.Cells[0,i+1]:=IntToStr(i+1);  //Wenn anfängliche Cloud-Länge =0, braucht man diese Zeile, um Nummerierung zu starten.
    while i<length(ref_Cloud) do begin
      SG_PEdit.Cells[0,i+1]:=IntToStr(i+1);
      SG_PEdit.Cells[1,i+1]:=FloatToStr(ref_Cloud[i].x);
      SG_PEdit.Cells[2,i+1]:=FloatToStr(ref_Cloud[i].y);
      inc(i,1);
    end;
    SG_PEdit.Cells[0,i+1]:=IntToStr(i+1);
    enableOnSelect:=true;
  end;

function TFPointEdit.GuiToPointString(forceSort:boolean):String;
 var
   errNotNumber,errEmpty:qword;  //errNotNumber=keine zahl in zelle; errEmpty=Zelle leer
   i:qword=1;
   z:qword;
   RowBuf,Buf:string;
   xtwice:boolean=false;
   xBuf:array of Double;
  begin
    result:='';
    errNotNumber:=0;
    errEmpty:=0;
    Buf:='';
    DeleteEmptyRows(true);
    setlength(xBuf,SG_PEdit.RowCount-1);
    while i<SG_PEdit.RowCount-1 do begin
      RowBuf:='';
      if (SG_PEdit.Cells[1,i]<>'')and(SG_PEdit.Cells[2,i]<>'') then begin
        try
          xBuf[i]:=StrToFloat(SG_PEdit.Cells[1,i]);
          z:=1;
          while z<i do begin
            if xBuf[z]=xBuf[i]then xTwice:=true;
            inc(z,1);
          end;
          RowBuf:=SG_PEdit.Cells[1,i]+'|'+FloatToStr(StrToFloat(SG_PEdit.Cells[2,i]))+';';
          Buf:=Buf+RowBuf;
        except
          inc(errNotNumber,1);
        end;
      end else if (SG_PEdit.Cells[1,i]='')xor(SG_PEdit.Cells[2,i]='') then inc(errEmpty,1);
      inc(i,1);
    end;
    If (errEmpty>0) or (errNotNumber>0) then
      if MessageDlg('*Eingabefehler*','Die Eingabe ist nicht einwandfrei. Es gibt '+IntToStr(errNotNumber)+' fehlerhafte Zahlen und '
                  +IntToStr(errEmpty)+' unvollstängige Wertepaare. Drücken Sie bitte auf "Yes" um fehlerhafte Einträge zu ignorieren.'
                  +'Drücken Sie "No" um die Einträge zu überarbeiten.', mtconfirmation, [mbYes,mbNo] , 0)=MrNo
                  then result:='korrektur';

    If (xTwice)and(CB_Ctype.ItemIndex>0) then begin
      if MessageDlg('*Eingabefehler*','Um Punkte verbinden zu können, dürfen nicht einem bestimmten x-Wert mehrere y-Werte zugewiesen werden.'
                     +'"Yes" drücken um Verbindungstyp auf "lose" zu setzen und fortzufahren oder "No" zum bearbeiten.',mtconfirmation,[mbYes,mbNo],0)=MrYes then begin
        CB_cType.ItemIndex:=0;
      end else result:='korrektur';
    end;
    if result='' then begin
      if (CB_ordered.checked)or(forceSort) then RowBuf:='s;' else RowBuf:='u;';
      case CB_ctype.ItemIndex of
        0:RowBuf:=RowBuf+'n';
        1:Rowbuf:=Rowbuf+'l';
      end;
      if Buf<>'' then result:='PointCloud('+buf+RowBuf+')'else Result:='';
    end;
    xBuf:=nil;
  end;


procedure TFPointEdit.SB_ReadyClick(Sender: TObject);
var
 strbuf:string;
begin
  enableOnSelect:=false;
  Strbuf:=GuiToPointString(false);
  If Strbuf<>'korrektur' then begin
    Case ServiceCaller of
      0: begin FMain.CB_Mainfunc.Text:=Strbuf; FMain.updateGraph;end;
      1: FFuncEdit.LE_FuncText.Text:=Strbuf;
      2: FEinstellungen.CB_PrevCurv.Text:=StrBuf;
    end;
    FPointEdit.ModalResult:=mrOK;              //schließt das Fenster ordnungsgemäß
  end else enableOnSelect:=true;
end;

procedure TFPointEdit.SG_PEditMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
  var
    col,row:longint;
begin
  If Button=mbright then begin
    enableOnSelect:=false;
    SG_Pedit.MouseToCell(x,y,col,Row);
    SG_PEdit.col:=col;
    SG_PEdit.Row:=row;
    enableOnSelect:=true;
  end;
end;

procedure TFPointEdit.SG_PEditSelectCell(Sender: TObject; aCol, aRow: Integer; var CanSelect: Boolean);
begin
  If (aRow=SG_PEdit.RowCount-1)and(enableOnSelect)and((SG_PEdit.cells[1,aRow]<>'')or(SG_PEdit.cells[2,aRow]<>'')) then begin
    SG_PEdit.RowCount:=aRow+2;
    SG_PEdit.Cells[0,aRow+1]:=IntToStr(aRow+1);
  end;
end;

procedure TFPointEdit.SpeedButton1Click(Sender: TObject);
begin
  enableOnSelect:=false;
  SG_PEdit.RowCount:=3;
  SG_PEdit.Col:=1; SG_Pedit.row:=1;  //Die Selection für das Grid auf 1,1 setzen
  SG_PEdit.Clean([gzNormal]);
  enableOnSelect:=true;
end;



procedure TFPointEdit.FormShow(Sender: TObject);
begin
  {$ifdef linux}
    FPointEdit.Font.Size:=9;
  {$endif}
  InitGui(Cloudstr,cloud);
end;

procedure TFPointEdit.PM_PEditClose(Sender: TObject);
begin
  SG_PEdit.Options:=SG_PEdit.Options-[goRowSelect];
end;

procedure TFPointEdit.PM_PEditPopup(Sender: TObject);
begin
  SG_PEdit.Options:=SG_PEdit.Options+[goRowSelect];
end;

procedure TFPointEdit.PM_PEdit_delRowClick(Sender: TObject);
begin
  SG_PEdit.DeleteColRow(false,SG_PEdit.Row);
  refreshCounter;
end;

procedure TFPointEdit.PM_PEdit_newRowClick(Sender: TObject);
begin
  SG_PEdit.InsertColRow(false,SG_PEdit.Row+1);
  refreshCounter;
end;

procedure TFPointEdit.PM_PEdit_sortViewClick(Sender: TObject);
 var
   loc_Cloud:tPointCloud;
   ordered,ctype:char;
   strbuf:string;
begin
  strbuf:=GuiToPointString(true);
  Numerik.StrToPCloud(strbuf,loc_Cloud,ordered,ctype);
  InitGui(strbuf,loc_Cloud);
  loc_Cloud:=nil;
end;

procedure TFPointEdit.FormClick(Sender: TObject);
begin
  deleteEmptyRows(false);
end;

procedure TFPointEdit.CB_orderedChange(Sender: TObject);
begin
  CB_ctype.Enabled:=cb_ordered.checked;
end;


initialization
  {$I upointedit.lrs}

end.

